#!/usr/bin/env python
'''
Created on 05-09-2013

@author: Damian Rusinek <damian.rusinek@gmail.com>
'''
import os
import wx
import wx.richtext
import wx.lib.newevent
import wx.lib.delayedresult

import aqopa

"""
from aqopa.gui.main_notebook_gui import MainNotebook, EVT_MODULE_SIMULATION_REQUEST, EVT_MODULE_SIMULATION_ALLOWED, EVT_MODULE_SIMULATION_FINISHED
from aqopa.gui.main_frame_gui import LibraryFrame, EVT_MODEL_SELECTED

"""
# AQoPA gui imports
from aqopa.gui.models_lib_gui import LibraryFrame, EVT_MODEL_SELECTED
from aqopa.gui.modules_panel_gui import ModulesPanel, EVT_MODULES_CHANGED
from aqopa.gui.mmv_panel_gui import MMVPanel
from aqopa.gui.run_panel_gui import RunPanel, EVT_MODEL_PARSED
from aqopa.gui.results_panel_gui import ResultsPanel

# modules communication events
ModuleSimulationRequestEvent, EVT_MODULE_SIMULATION_REQUEST = wx.lib.newevent.NewEvent() # Parameters: module
ModuleSimulationAllowedEvent, EVT_MODULE_SIMULATION_ALLOWED = wx.lib.newevent.NewEvent() # Parameters: interpreter
ModuleSimulationFinishedEvent, EVT_MODULE_SIMULATION_FINISHED = wx.lib.newevent.NewEvent()

class MainNotebook(wx.Notebook):
    """ """

    def __init__(self, parent):
        wx.Notebook.__init__(self, parent)

        ###########
        # MODULES
        ###########


        # here you can add modules to the GUI version of AQoPA
        self.availableModules = []

        # add time analysis module
        from aqopa.module import timeanalysis
        timeanalysis_module = timeanalysis.Module()
        timeanalysis_module.get_gui().Bind(EVT_MODULE_SIMULATION_REQUEST,
                                           self.OnModuleSimulationRequest)
        timeanalysis_module.get_gui().Bind(EVT_MODULE_SIMULATION_FINISHED,
                                           self.OnModuleSimulationFinished)
        self.availableModules.append(timeanalysis_module)

        # add energy analysis module - it depends on time analysis module
        from aqopa.module import energyanalysis
        energyanalysis_module = energyanalysis.Module(timeanalysis_module)
        self.availableModules.append(energyanalysis_module)

        # add reputation module
        from aqopa.module import reputation
        self.availableModules.append(reputation.Module())

        # add qop module - KM
        from aqopa.module import qopanalysis
        self.availableModules.append(qopanalysis.Module())

        # add finance module - it depends on energy analysis module - KM
        from aqopa.module import financialanalysis
        fm = financialanalysis.Module(energyanalysis_module)
        self.availableModules.append(fm)

        # add gogreen! module - it depends on energy analysis module - KM
        from aqopa.module import greenanalysis
        gm = greenanalysis.Module(energyanalysis_module)
        self.availableModules.append(gm)

        # list containing notebook images:
        # .ico seem to be more OS portable, although we use .png here
        # the (20, 20) is the size in pixels of the images
        il = wx.ImageList(20, 20)
        modelsTabImg = il.Add(wx.Bitmap(self.CreatePath4Resource('models_lib.png'), wx.BITMAP_TYPE_PNG))
        metricsTabImg = il.Add(wx.Bitmap(self.CreatePath4Resource('metrics.png'), wx.BITMAP_TYPE_PNG))
        versionsTabImg = il.Add(wx.Bitmap(self.CreatePath4Resource('versions.png'), wx.BITMAP_TYPE_PNG))
        runTabImg = il.Add(wx.Bitmap(self.CreatePath4Resource('run.png'), wx.BITMAP_TYPE_PNG))
        modulesTabImg = il.Add(wx.Bitmap(self.CreatePath4Resource('modules.png'), wx.BITMAP_TYPE_PNG))
        resultsTabImg = il.Add(wx.Bitmap(self.CreatePath4Resource('results.png'), wx.BITMAP_TYPE_PNG))
        self.AssignImageList(il)

        ###########
        # TABS
        ###########

        self.modelTab = MMVPanel(self)
        self.modelTab.Layout()
        self.Bind(wx.EVT_TEXT, self.OnModelTextChange, self.modelTab.dataTextArea)
        self.AddPage(self.modelTab, "Model")
        self.SetPageImage(0, modelsTabImg)

        self.metricsTab = MMVPanel(self)
        self.metricsTab.Layout()
        self.Bind(wx.EVT_TEXT, self.OnModelTextChange, self.metricsTab.dataTextArea)
        self.AddPage(self.metricsTab, "Metrics")
        self.SetPageImage(1, metricsTabImg)

        self.versionsTab = MMVPanel(self)
        self.versionsTab.Layout()
        self.Bind(wx.EVT_TEXT, self.OnModelTextChange, self.versionsTab.dataTextArea)
        self.versionsTab.Layout()
        self.AddPage(self.versionsTab, "Versions")
        self.SetPageImage(2, versionsTabImg)

        self.modulesTab = ModulesPanel(self, modules=self.availableModules)
        self.modulesTab.Bind(EVT_MODULES_CHANGED, self.OnModulesChange)
        self.modulesTab.Layout()
        self.AddPage(self.modulesTab, "Modules")
        self.SetPageImage(3, modulesTabImg)

        self.runTab = RunPanel(self)
        self.runTab.SetAllModules(self.availableModules)
        self.runTab.Layout()
        self.runTab.Bind(EVT_MODEL_PARSED, self.OnModelParsed)
        self.AddPage(self.runTab, "Run")
        self.SetPageImage(4, runTabImg)

        self.resultsTab = ResultsPanel(self)
        self.resultsTab.Layout()
        self.AddPage(self.resultsTab, "Results")
        self.SetPageImage(5, resultsTabImg)

    def CreatePath4Resource(self, resourceName):
        """
        @brief      creates and returns path to the
                    given file in the resource
                    ('assets') dir
        @return     path to the resource
        """
        tmp = os.path.split(os.path.dirname(__file__))
        return os.path.join(tmp[0], 'bin', 'assets', resourceName)

    def LoadModelFile(self, filePath):
        self.modelTab.dataTextArea.LoadFile(filePath)

    def LoadMetricsFile(self, filePath):
        self.metricsTab.dataTextArea.LoadFile(filePath)

    def LoadVersionsFile(self, filePath):
        self.versionsTab.dataTextArea.LoadFile(filePath)

    def SetModelData(self, data):
        self.modelTab.dataTextArea.SetValue(data)

    def SetMetricsData(self, data):
        self.metricsTab.dataTextArea.SetValue(data)

    def SetVersionsData(self, data):
        self.versionsTab.dataTextArea.SetValue(data)

    def GetModelData(self):
        return self.modelTab.dataTextArea.GetValue().strip()

    def GetMetricsData(self):
        return self.metricsTab.dataTextArea.GetValue().strip()

    def GetVersionsData(self):
        return self.versionsTab.dataTextArea.GetValue().strip()

    def OnModelTextChange(self, event):
        self.runTab.SetModel(self.GetModelData(),
                             self.GetMetricsData(),
                             self.GetVersionsData())
        event.Skip()

    def OnModulesChange(self, event):
        self.runTab.SetSelectedModules(event.modules)
        self.resultsTab.SetSelectedModules(event.modules)

    def OnModelParsed(self, event):
        self.resultsTab.ClearResults()
        event.Skip()

    def OnModuleSimulationRequest(self, event):
        """ """
        gui = event.module.get_gui()
        self.runTab.runButton.Enable(False)
        self.runTab.parseButton.Enable(False)

        wx.PostEvent(gui, ModuleSimulationAllowedEvent(interpreter=self.runTab.interpreter))

    def OnModuleSimulationFinished(self, event):
        """ """
        self.runTab.parseButton.Enable(True)

class MainFrame(wx.Frame):
    """ """
    def __init__(self, *args, **kwargs):
        wx.Frame.__init__(self, *args, **kwargs)

        ###########
        # MENUBAR
        ###########

        # create menubar
        menuBar = wx.MenuBar()

        #create main menu, lets call it 'file' menu
        fileMenu = wx.Menu()
        # create menu item = about AQoPA
        item = wx.MenuItem(fileMenu, wx.NewId(), u"&About AQoPA\tCTRL+I")
        item.SetBitmap(wx.Bitmap(self.CreatePath4Resource('about.png')))
        fileMenu.AppendItem(item)
        self.Bind(wx.EVT_MENU, self.OnAbout, item)
        fileMenu.AppendSeparator()
        # create menu item = quit AQoPa
        item = wx.MenuItem(fileMenu, wx.NewId(), u"&Quit\tCTRL+Q")
        item.SetBitmap(wx.Bitmap(self.CreatePath4Resource('exit.png')))
        fileMenu.AppendItem(item)
        self.Bind(wx.EVT_MENU, self.OnQuit, item)
        # add 'file' menu to the menubar
        menuBar.Append(fileMenu, "&Menu")

        # create library menu, here u can find modules library
        libraryMenu = wx.Menu()
        # create models menu item
        item = wx.MenuItem(libraryMenu, wx.NewId(), u"&Browse models\tCTRL+M")
        item.SetBitmap(wx.Bitmap(self.CreatePath4Resource('lib.png')))
        libraryMenu.AppendItem(item)
        self.Bind(wx.EVT_MENU, self.OnBrowseModels, item)

        # create metric menu item
        # item = wx.MenuItem(libraryMenu, wx.NewId(), u"&Browse metrics\tCTRL+F")
        # item.SetBitmap(wx.Bitmap(self.CreatePath4Resource('metrics.png')))
        # libraryMenu.AppendItem(item)

        # add 'library' menu to the menubar
        menuBar.Append(libraryMenu, "&Library")

        self.SetMenuBar(menuBar)

        ###################
        # SIZERS & EVENTS
        ###################

        self.mainNotebook = MainNotebook(self)

        logoPanel = wx.Panel(self)
        pic = wx.StaticBitmap(logoPanel)
        pic.SetBitmap(wx.Bitmap(self.CreatePath4Resource('logo.png')))

        sizer = wx.BoxSizer(wx.HORIZONTAL)
        sizer.Add(wx.StaticText(self), 1, wx.EXPAND, 5)
        sizer.Add(logoPanel, 0, wx.RIGHT| wx.ALL|wx.EXPAND, 5)
        s2 = wx.BoxSizer(wx.VERTICAL)
        s2.Add(sizer, 0, wx.LEFT| wx.ALL|wx.EXPAND, 5)
        s2.Add(self.mainNotebook, 1, wx.ALL|wx.EXPAND, 5)
        self.SetSizer(s2)

        self.SetIcon(wx.Icon(self.CreatePath4Resource('app_logo.png'), wx.BITMAP_TYPE_PNG))
        # the size of the whole application window
        self.SetClientSize(wx.Size(800, 600))
        self.CenterOnScreen()
        self.Layout()

    def CreatePath4Resource(self, resourceName):
        """
        @brief      creates and returns path to the
                    given file in the resource
                    ('assets') dir
        @return     path to the resource
        """
        tmp = os.path.split(os.path.dirname(__file__))
        return os.path.join(tmp[0], 'bin', 'assets', resourceName)

    def OnQuit(self, event=None):
        """
        @brief  closes the application
        """
        self.Close()

    def OnBrowseModels(self, event=None):
        """
        @brief  shows the library frame (models library window)
        """
        libraryFrame = LibraryFrame(self, title="Models Library")
        libraryFrame.Show(True)
        libraryFrame.CentreOnParent()
        #libraryFrame.Maximize(True)
        libraryFrame.Bind(EVT_MODEL_SELECTED, self.OnLibraryModelSelected)

    def OnLibraryModelSelected(self, event):
        """ """
        self.mainNotebook.SetModelData(event.model_data)
        self.mainNotebook.SetMetricsData(event.metrics_data)
        self.mainNotebook.SetVersionsData(event.versions_data)
        # set filenames on GUI
        self.mainNotebook.modelTab.SetFilenameOnGUI(event.filenames['model'])
        self.mainNotebook.metricsTab.SetFilenameOnGUI(event.filenames['metrics'])
        self.mainNotebook.versionsTab.SetFilenameOnGUI(event.filenames['versions'])

    def OnAbout(self, event=None):
        """ Show about info """

        description = """AQoPA stands for Automated Quality of Protection Analysis Tool
        for QoPML models."""

        licence = """AQoPA is free software; you can redistribute
it and/or modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of the License,
or (at your option) any later version.

AQoPA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE."""

        logo_filepath = self.CreatePath4Resource('logo.png')

        info = wx.AboutDialogInfo()
        info.SetIcon(wx.Icon(logo_filepath, wx.BITMAP_TYPE_PNG))
        info.SetName('AQoPA')
        info.SetVersion(aqopa.VERSION)
        info.SetDescription(description)
        info.SetCopyright('(C) 2013 QoPML Project')
        info.SetWebSite('http://www.qopml.org')
        info.SetLicence(licence)
        info.AddDeveloper('Damian Rusinek')
        info.AddDocWriter('Damian Rusinek')
        info.AddArtist('QoPML Project')
        info.AddTranslator('Damian Rusinek')

        wx.AboutBox(info)

class AqopaApp(wx.App):

    def OnInit(self):
        self.mainFrame = MainFrame(None,
                                   title="Automated Quality of Protection Analysis Tool")
        self.mainFrame.Show(True)
        self.mainFrame.CenterOnScreen()
       # self.mainFrame.Maximize(True)
        self.SetTopWindow(self.mainFrame)

        return True
    




'''
Created on 07-05-2013

@author: Damian Rusinek <damian.rusinek@gmail.com>
'''

import wx
import wx.lib.newevent
import threading
from aqopa.model.parser import ParserException, ModelParserException,\
    MetricsParserException, ConfigurationParserException
from aqopa.model.parser.lex_yacc.grammar import algorithms
from aqopa.model.store import QoPMLModelStore
from aqopa.model import HostProcess, name_indexes, original_name,\
    HostSubprocess, WhileInstruction, IfInstruction
from aqopa.simulator import Simulator,\
    expression, state, equation, metrics, communication, scheduler, predefined, algorithm

from aqopa.simulator.state import Executor,\
    AssignmentInstructionExecutor, IfInstructionExecutor,\
    ProcessInstructionExecutor, SubprocessInstructionExecutor,\
    FinishInstructionExecutor, CommunicationInstructionExecutor,\
    ContinueInstructionExecutor, WhileInstructionExecutor, Host, Process,\
    CallFunctionInstructionExecutor, PrintExecutor, BreakInstructionExecutor

from aqopa.simulator.error import EnvironmentDefinitionException

class Builder():
    """
    Builder that builds environment elements
    """
    
    def build_store(self):
        """
        Builds store - keeps all parsed elements from qopml model
        """
        return QoPMLModelStore()
    
    def _build_functions(self, store):
        """
        Validate and build functions
        """
        functions = []
        errors = []
        
        for f in store.functions:
            err = False
            for ef in functions:
                if ef.name == f.name:
                    errors.append("Function %s redeclared" % f.name)
                    err = True
                    break
            if not err:
                functions.append(f)
                
        if len(errors) > 0:
            raise EnvironmentDefinitionException('Functions redeclaration.', errors=errors)
        
        return functions
    
    def _build_equations(self, store, functions):
        """
        Validate, build and return simulation equations build from parsed equations.
        """
        validator = equation.Validator()
        validator.validate(store.equations, functions)
        
        equations = []
        for parsed_equation in store.equations:
            equations.append(equation.Equation(parsed_equation.composite, parsed_equation.simple))
        return equations
    
    def _build_hosts(self, store, version, functions, channels, populator, reducer):
        """
        Rebuild hosts - create new instances with updated instructions lists according to version.
        """
        
        def remove_unused_subprocesses(instructions_list, run_process): 
            """
            Create instruction lists according to "run process"
            Removes subprocesses that are not selected to run. 
            """
            if run_process.all_subprocesses_active:
                return instructions_list
            
            new_instructions_list = []
            
            for instruction in instructions_list:
                if isinstance(instruction, HostSubprocess):
                    subprocess = instruction
                    if subprocess.name in run_process.active_subprocesses:
                        new_instructions_list.append(subprocess)
                        subprocess.instructions_list = \
                            remove_unused_subprocesses(subprocess.instructions_list, 
                                                       run_process)
                else:
                    new_instructions_list.append(instruction)
                    
                    if isinstance(instruction, WhileInstruction):
                        instruction.instructions = \
                            remove_unused_subprocesses(instruction.instructions, 
                                                       run_process)
                    
                    if isinstance(instruction, IfInstruction):
                        instruction.true_instructions = \
                            remove_unused_subprocesses(instruction.true_instructions, 
                                                       run_process)
                        instruction.false_instructions = \
                            remove_unused_subprocesses(instruction.false_instructions, 
                                                       run_process)
            
            return new_instructions_list
            
        def get_channels_assgined_to_host(run_host, parsed_host, built_channels):
            """
            Creates list all channels that host can use.
            Firstly it is checked in version, secondly in parsed host.
            """
            channel_names = []
            if run_host.all_channels_active:
                if parsed_host.all_channels_active:
                    return built_channels
                else:
                    channel_names.extend([ c for c in parsed_host.active_channels ])
            else:
                channel_names.extend([ c for c in run_host.active_channels ])
            # channel_names - names of channels that host can use
            channels = [] 
            for channel_name in channel_names:
                for channel in built_channels:
                    if channel.name == channel_name:
                        channels.append(channel)
            return channels
        
        def get_channels_assigned_to_process(run_process, parsed_process, built_channels):
            """
            Creates list all channels that process can use.
            Firstly it is checked in version, secondly in parsed process.
            """
            channel_names = []
            # Get channel names uded by process
            if parsed_process.all_channels_active:
                return built_channels
            else:
                for channel_name in parsed_process.active_channels:
                    channel_names.append(channel_name)
            # channel_names - names of channels that host can use
            channels = [] 
            for channel_name in channel_names:
                for channel in built_channels:
                    if channel.name == channel_name:
                        channels.append(channel)
            return channels
            
        def build_host_instructions_list(parsed_host, run_host, repetition_number, channels):
            """
            Create host's  and its processes' instruction lists 
            according to "run host" (repetitions) 
            """
            
            def find_process(instructions_list, process_name):
                for instr in instructions_list:
                    if not isinstance(instr, HostProcess):
                        continue
                    if instr.name == process_name:
                        return instr
                return None
            
            
            processes_numbers = {}
            host_instructions_list = []
            
            for run_process in run_host.run_processes:
                
                # Find parsed process
                parsed_process = find_process(parsed_host.instructions_list, run_process.process_name)
                
                if parsed_process is None:
                    raise EnvironmentDefinitionException("Process '%s' does not exist in host '%s'." % 
                                                         (run_process.process_name, parsed_host.name))
                
                # Clone the parsed process, because builder 
                # may change its the instructions list
                # and remove some subprocesses that would be 
                # needed by other version 
                parsed_process = parsed_process.clone()
                
                # Define initial process number (if needed)
                if run_process.process_name not in processes_numbers:
                    processes_numbers[run_process.process_name] = 0
                
                # If process has follower
                if run_process.follower:
                    # Define initial process number for following process (if needed)
                    if run_process.follower.process_name not in processes_numbers:
                        processes_numbers[run_process.follower.process_name] = 0
                        
                # Get channels assigned to process and its follower (if needed)
                process_channels = get_channels_assigned_to_process(run_process, parsed_process, channels)
                follower_channels = []
                if run_process.follower:
                    follower_parsed_process = find_process(parsed_host.instructions_list, run_process.follower.process_name)
                    follower_channels = get_channels_assigned_to_process(run_process.follower, follower_parsed_process, channels)
                    
                for i in range(0, run_process.repetitions):
                    
                    # Create instructions list
                    instructions_list = remove_unused_subprocesses(parsed_process.instructions_list, run_process)
                    
                    # Create new simulation process
                    simulated_process = Process(parsed_process.name, instructions_list)

                    # Connect with channels
                    for ch in process_channels:
                        ch.connect_with_process(simulated_process)
                    
                    # Update process index
                    process_number = processes_numbers[run_process.process_name] + i
                    simulated_process.add_name_index(process_number)
                    
                    # Do the same fo the follower 
                    if run_process.follower:
                        
                        # Find process of follower
                        follower_parsed_process = find_process(parsed_host.instructions_list, run_process.follower.process_name)

                        # Build instructions list for follower
                        follower_instructions_list = remove_unused_subprocesses(follower_parsed_process.instructions_list, run_process.follower)

                        # Create simulated follower
                        simulated_follower = Process(follower_parsed_process.name, follower_instructions_list)

                        # Connect with channels
                        for ch in follower_channels:
                            ch.connect_with_process(simulated_follower)
                            
                        # Update follower index
                        follower_number = processes_numbers[run_process.follower.process_name] + i
                        simulated_follower.add_name_index(follower_number)
                        
                        simulated_process.follower = simulated_follower
                        
                    host_instructions_list.append(simulated_process)
            
            return host_instructions_list
        
        def set_scheduler(host, algorithm):
            """
            Build and set host's scheduler
            """
            host.set_scheduler(scheduler.create(host, algorithm))
        
        built_hosts = []
            
        # Hosts numbers dict keeps the last used number of repeated host.
        hosts_numbers = {} 
        
        for run_host in version.run_hosts:
            
            if run_host.host_name not in hosts_numbers:
                hosts_numbers[run_host.host_name] = 0
            
            # Create prototype parsed host for this "run host"
            parsed_host = store.find_host(run_host.host_name)
            
            assigned_channels = get_channels_assgined_to_host(run_host, parsed_host, channels)
            
            for i in range(0, run_host.repetitions):

                # Build next instructions list for next repeated host                
                instructions_list = build_host_instructions_list(parsed_host, run_host, i, channels)
                
                simulation_host = Host(parsed_host.name, instructions_list)
                    
                # Set the number of host    
                host_number = hosts_numbers[run_host.host_name] + i
                simulation_host.add_name_index(host_number)
                
                # Set scheduler
                set_scheduler(simulation_host, parsed_host.schedule_algorithm)
                
                for ch in assigned_channels:
                    ch.connect_with_host(simulation_host)
                
                built_hosts.append(simulation_host)
                
            hosts_numbers[run_host.host_name] += run_host.repetitions 
            
        return built_hosts

    def _set_hosts_predefined_values(self, store, hosts, populator):
            
        def set_predefined_variables(host, predefined_values, populator):
            """
            Populate predefined values with expressions 
            and save them as variables in host
            """
            for predefined_value in predefined_values:
                populated_value = populator.populate(predefined_value.expression.clone(), host)
                # print 'Setting predefined variable ', predefined_value.variable_name, \
                #     ' in host ', host.name, ' with value ', unicode(populated_value), \
                #     ' (', getattr(populated_value, '_host_name', 'None'), ')'
                host.set_variable(predefined_value.variable_name, populated_value)
        for host in hosts:
            parsed_host = store.find_host(host.original_name())
            # Save predefined values as variables
            set_predefined_variables(host, parsed_host.predefined_values, populator)
        return hosts
    
    def _build_expression_populator(self, reducer):
        """
        Build and return object that populates expressions
        with variables' values.
        """
        return expression.Populator(reducer)
    
    def _build_expression_checker(self, populator):
        """
        Build and return object that checks the logic value of expressions.
        """
        return expression.Checker(populator)
    
    def _build_expression_reducer(self, equations):
        """
        Build and return object that reduces expressions.
        """
        return expression.Reducer(equations)
    
    def _build_channels(self, store, version):
        """
        Validate, build and return simulation channels build from parsed channels.
        Includes channel repetitions.
        """
        
        built_channels = []
        for parsed_channel in store.channels:
            channel = communication.Channel(parsed_channel.name, parsed_channel.buffor_size, parsed_channel.tag_name)
            built_channels.append(channel)
        return built_channels
    
    def _build_topology(self, topology_rules, hosts):

        def find_left_hosts(topology_host, hosts):
            found_hosts = []
            for host in hosts:
                # If host has the same identifier
                if host.original_name() == topology_host.identifier:
                    # If no range is specified
                    if topology_host.index_range is None:
                        found_hosts.append(host)
                    else:
                        # Range is specified
                        start_i = topology_host.index_range[0]
                        end_i = topology_host.index_range[1]
                        i = host.get_name_index()
                        if (start_i is None or i >= start_i) and (end_i is None or i <= end_i):
                            found_hosts.append(host)
            return found_hosts
        
        def find_right_hosts(topology_host, hosts, current_host):
            if topology_host is None:
                return []
            found_hosts = []
            for host in hosts:
                # If host has the same identifier
                if host.original_name() == topology_host.identifier:
                    # If no range is specified
                    if topology_host.index_range is None:
                        # If no index shift is specified
                        if topology_host.i_shift is None:
                            found_hosts.append(host)
                        else:
                            # Index shift is specified
                            i = current_host.get_name_index()
                            i += topology_host.i_shift
                            shifted_host_name = topology_host.identifier + "." + str(i)
                            if host.name == shifted_host_name:
                                found_hosts.append(host)
                    else:
                        # Range is specified
                        start_i = topology_host.index_range[0]
                        end_i = topology_host.index_range[1]
                        i = host.get_name_index()
                        if (start_i is None or i >= start_i) and (end_i is None or i <= end_i):
                            found_hosts.append(host)
            return found_hosts
        
        def add_connection(topology, from_host, to_host, parameters):
            if from_host not in topology:
                topology[from_host] = {'hosts': [], 'parameters': {}}
            if to_host not in topology[from_host]['hosts']:
                topology[from_host]['hosts'].append(to_host)
                for parameter in parameters:
                    if parameter not in topology[from_host]['parameters']:
                        topology[from_host]['parameters'][parameter] = {}
                    topology[from_host]['parameters'][parameter][to_host] = parameters[parameter]
            return topology

        topology = {}
        for rule in topology_rules:
            for left_host in find_left_hosts(rule.left_host, hosts):
                for right_host in find_right_hosts(rule.right_host, hosts, left_host):
                    if rule.arrow == '->' or rule.arrow == '<->':
                        topology = add_connection(topology, left_host, right_host, rule.parameters)
                    if rule.arrow == '<-' or rule.arrow == '<->':
                        topology = add_connection(topology, right_host, left_host, rule.parameters)
        return topology
        
    
    def _build_channels_manager(self, channels, built_hosts, version, store):
        """ 
        Build channels manager
        """
        mgr = communication.Manager(channels)
        for name in version.communication['mediums']:
            topology_rules = version.communication['mediums'][name]['topology']['rules']
            default_params = version.communication['mediums'][name]['default_parameters']
            mgr.add_medium(name, self._build_topology(topology_rules, built_hosts), default_params)
        for name in store.mediums:
            if not mgr.has_medium(name):
                topology_rules = store.mediums[name]['topology']['rules']
                default_params = store.mediums[name]['default_parameters']
                mgr.add_medium(name, self._build_topology(topology_rules, built_hosts), default_params)
        return mgr
    
    def _build_predefined_functions_manager(self, context):
        """
        Build manager for predefined functions
        """
        return predefined.FunctionsManager(context)
        
    def _build_metrics_manager(self, store, hosts, version):
        """
        Build and return metrics manager.
        """
        host_metrics = []
        
        for metrics_data in store.metrics_datas:
            
            blocks = []
            # Build list of metrics blocks
            for block in metrics_data.blocks:
                
                params = block.header.params[1:]
                service_params = block.header.services_params
                metrics_block = metrics.Block(params, service_params)
                
                for metric in block.metrics:
                    m = metrics.Metric(metric.arguments[0], metric.arguments[1:len(params)+1], 
                                       metric.arguments[len(params)+1:])
                    metrics_block.add_metric(m)
                blocks.append(metrics_block)
            
            # get or create host metrics with given name
            hm = None
            for existing_hm in host_metrics:
                if existing_hm.name == metrics_data.name:
                    hm = existing_hm
                    break
            if hm is None:
                hm = metrics.HostMetrics(metrics_data.name)
                host_metrics.append(hm)
            
            if metrics_data.plus or metrics_data.star:
                if metrics_data.plus:
                    hm.plus_blocks = blocks
                else: # star
                    hm.star_blocks = blocks
                    
                # If host metrics does not have normal block
                # Search for them and assign if found in host metrics 
                # with simple name (not qualified like hm1.1)
                if len(hm.normal_blocks) == 0:
                    hm_original_name = original_name(hm.name)
                    hm_original = None
                    
                    for existing_hm in host_metrics:
                        if original_name(existing_hm.name) == hm_original_name \
                            and existing_hm != hm:
                            hm_original = existing_hm
                            break
                        
                    if hm_original:
                        hm.normal_blocks = hm_original.normal_blocks
                        
            else: # metrics_data normal
                hm_original_name = original_name(hm.name)
                
                # Assign normal block to all host metrics with the same original name
                for existing_hm in host_metrics:
                    if original_name(existing_hm.name) == hm_original_name:
                        # Assign notmal block to all host metrics with the same original name
                        # (including the current one - possibly created)
                        existing_hm.normal_blocks = blocks
                        
        # Connect host metrics with hosts
        for metrics_set in version.metrics_sets:
            for h in hosts:
                if h.original_name() == metrics_set.host_name:
                    for host_metric in host_metrics:
                        if host_metric.name == metrics_set.configuration_name:
                            host_metric.connected_hosts.append(h)
                            break
        
        return metrics.Manager(host_metrics)

    def _build_algorithms_resolver(self, store):
        """
        """
        resolver = algorithm.AlgorithmResolver()
        for alg_name in store.algorithms:
            resolver.add_algorithm(alg_name, store.algorithms[alg_name])
        return resolver
    
    def _build_context(self, store, version):
        """
        Builds context with initial state.
        """
        functions = self._build_functions(store)
        equations = self._build_equations(store, functions)
        expression_reducer = self._build_expression_reducer(equations)
        expression_populator = self._build_expression_populator(expression_reducer)
        expression_checker = self._build_expression_checker(expression_populator)
        channels = self._build_channels(store, version)
        hosts = self._build_hosts(store, version, functions, channels,
                                  expression_populator, expression_reducer)

        # Context
        c = state.Context(version)
        c.functions = functions
        c.hosts = hosts
        c.expression_reducer = expression_reducer
        c.expression_checker = expression_checker
        c.expression_populator = expression_populator
        c.metrics_manager = self._build_metrics_manager(store, hosts, version)
        c.channels_manager = self._build_channels_manager(channels, hosts, version, store)
        c.algorithms_resolver = self._build_algorithms_resolver(store)
        
        # Predefined manager
        predefined_functions_manager = self._build_predefined_functions_manager(c)
        expression_populator.predefined_functions_manager = predefined_functions_manager
        expression_reducer.predefined_functions_manager = predefined_functions_manager
        
        # Predefined hosts' variables
        self._set_hosts_predefined_values(store, hosts, expression_populator)
        
        return c
    
    def build_executor(self):
        """
        Creates executor for simulation
        """
        e = Executor()
        e.append_instruction_executor(AssignmentInstructionExecutor())
        e.append_instruction_executor(CallFunctionInstructionExecutor())
        e.append_instruction_executor(ProcessInstructionExecutor())
        e.append_instruction_executor(SubprocessInstructionExecutor())
        e.append_instruction_executor(CommunicationInstructionExecutor())
        e.append_instruction_executor(FinishInstructionExecutor())
        e.append_instruction_executor(ContinueInstructionExecutor())
        e.append_instruction_executor(BreakInstructionExecutor())
        e.append_instruction_executor(IfInstructionExecutor())
        e.append_instruction_executor(WhileInstructionExecutor())
        return e
    
    def build_simulator(self, store, version):
        """
        Creates simulator for particular version.
        """
        sim = Simulator(self._build_context(store, version))
        sim.set_executor(self.build_executor())
        return sim
    
    def build_model_parser(self, store, modules):
        """
        Builder parser that parses model written in QoPML
        and populates the store.
        """
        from aqopa.model.parser.lex_yacc import LexYaccParser
        from aqopa.model.parser.lex_yacc.grammar import main,\
                functions, channels, equations, expressions, instructions,\
                hosts, modules as modules_module, communication as comm_grammar
        
        parser = LexYaccParser()
        parser.set_store(store) \
                .add_extension(main.ModelParserExtension()) \
                .add_extension(modules_module.ModelParserExtension()) \
                .add_extension(functions.ModelParserExtension()) \
                .add_extension(channels.ModelParserExtension()) \
                .add_extension(equations.ModelParserExtension()) \
                .add_extension(expressions.ModelParserExtension()) \
                .add_extension(comm_grammar.ModelParserExtension()) \
                .add_extension(algorithms.ModelParserExtension()) \
                .add_extension(instructions.ModelParserExtension()) \
                .add_extension(hosts.ModelParserExtension())
                
        for m in modules:
            parser = m.extend_model_parser(parser)
                
        return parser.build()
    
    def build_metrics_parser(self, store, modules):
        """
        Builder parser that parses metrics written in QoPML
        and populates the store.
        """
        from aqopa.model.parser.lex_yacc import LexYaccParser
        from aqopa.model.parser.lex_yacc.grammar import main, metrics as metrics_grammar
        
        parser = LexYaccParser()
        parser.set_store(store)\
                .add_extension(main.MetricsParserExtension())\
                .add_extension(metrics_grammar.MetricsParserExtension())
                
        for m in modules:
            parser = m.extend_metrics_parser(parser)
                
        return parser.build()
    
    def build_config_parser(self, store, modules):
        """
        Builder parser that parses config written in QoPML
        and populates the store.
        """
        from aqopa.model.parser.lex_yacc import LexYaccParser
        from aqopa.model.parser.lex_yacc.grammar import versions, main, communication as comm_grammar
        
        parser = LexYaccParser()
        parser.set_store(store)\
                .add_extension(main.ConfigParserExtension())\
                .add_extension(comm_grammar.ConfigParserExtension()) \
                .add_extension(versions.ConfigParserExtension())
                
        for m in modules:
            parser = m.extend_config_parser(parser)
                
        return parser.build()
    
class Interpreter():
    """
    Interpreter is responsible for parsing the model,
    creating the environment for simulations,
    manipulating selected models.
    """
    
    def __init__(self, builder=None, model_as_text="", 
                 metrics_as_text="", config_as_text=""):
        self.builder = builder if builder is not None else Builder()
        self.model_as_text = model_as_text
        self.metrics_as_text = metrics_as_text 
        self.config_as_text = config_as_text
        
        self.store = None
        
        self._modules = []
        
    def set_qopml_model(self, model_as_text):
        """
        Set qopml model that will be interpreted.
        """
        self.model_as_text = model_as_text
        return self
    
    def set_qopml_metrics(self, metrics_as_text):
        """
        Set qopml metrics that will be interpreted.
        """
        self.metrics_as_text = metrics_as_text
        return self
    
    def set_qopml_config(self, config_as_text):
        """
        Set qopml configuration that will be interpreted.
        """
        self.config_as_text = config_as_text
        return self
    
    def register_qopml_module(self, qopml_module):
        """
        Registers new module
        """
        if qopml_module in self._modules:
            raise EnvironmentDefinitionException(u"QoPML Module '%s' is already registered." % unicode(qopml_module))
        self._modules.append(qopml_module)
        return self
        
    def parse(self, all_modules):
        """
        Parses the model from model_as_text field and populates the store.
        """
        if len(self.model_as_text) == 0:
            raise EnvironmentDefinitionException("QoPML Model not provided.")
    
        self.store = self.builder.build_store()
    
        parser = self.builder.build_model_parser(self.store, all_modules)
        parser.parse(self.model_as_text)
        if len(parser.get_syntax_errors()) > 0:
            raise ModelParserException('Invalid syntax.', syntax_errors=parser.get_syntax_errors())
    
        parser = self.builder.build_metrics_parser(self.store, all_modules)
        parser.parse(self.metrics_as_text)
        if len(parser.get_syntax_errors()) > 0:
            raise MetricsParserException('Invalid syntax.', syntax_errors=parser.get_syntax_errors())
    
        parser = self.builder.build_config_parser(self.store, all_modules)
        parser.parse(self.config_as_text)
        if len(parser.get_syntax_errors()) > 0:
            raise ConfigurationParserException('Invalid syntax.', syntax_errors=parser.get_syntax_errors())
    
    def install_modules(self, simulator):
        """ """
        raise NotImplementedError()
    
    def run(self):
        """ Runs all simulations """
        raise NotImplementedError()
            
class ConsoleInterpreter(Interpreter):
    
    def __init__(self, builder=None, model_as_text="", 
                 metrics_as_text="", config_as_text=""):
        Interpreter.__init__(self, builder, model_as_text, metrics_as_text, config_as_text)
        
        self.simulators = []

    def save_states_to_file(self, simulator):
        """ 
        Tells simulator to save states flow to file.
        """
        f = open('VERSION_%s_STATES_FLOW' % simulator.context.version.name, 'w')
        simulator.get_executor().prepend_instruction_executor(PrintExecutor(f))      

    def prepare(self):
        """ Prepares for run """
        
        for version in self.store.versions:
            simulator = self.builder.build_simulator(self.store, version)
            self.simulators.append(simulator)
            self.install_modules(simulator)
    
    def install_modules(self, simulator):
        """ """
        for m in self._modules:
            m.install_console(simulator)
        
    def run(self):
        """ Runs all simulations """

        for s in self.simulators:
            s.prepare()
            s.run()
            self.on_finished(s)
            
    def on_finished(self, simulator):
        pass

    def is_finished(self):
        for s in self.simulators:
            if not s.is_simulation_finished():
                return False
        return True

class GuiInterpreter(Interpreter):
    
    def __init__(self, builder=None, model_as_text="", 
                 metrics_as_text="", config_as_text=""):
        Interpreter.__init__(self, builder=builder, 
                             model_as_text=model_as_text, 
                             metrics_as_text=metrics_as_text, 
                             config_as_text=config_as_text)
        self.simulators = []

    def prepare(self):
        """ Prepares for run """
        for version in self.store.versions:
            simulator = self.builder.build_simulator(self.store, version)
            self.install_modules(simulator)
            self.simulators.append(simulator)
    
    def install_modules(self, simulator):
        """ """
        for m in self._modules:
            m.install_gui(simulator)
            
    def run_simulation(self, simulator):
        """ """
        simulator.prepare()
        simulator.run()
        return simulator



'''
Created on 22-04-2013

@author: Damian Rusinek <damian.rusinek@gmail.com>
'''

from aqopa.model.parser.lex_yacc import LexYaccParserExtension
from aqopa.model import Channel


class Builder():
    """
    Builder for creating channel objects
    """
    
    def build_channels(self, token):
        """
        channel : CHANNEL identifiers_list LPARAN channel_buffor RPARAN SEMICOLON
            | CHANNEL identifiers_list LPARAN channel_buffor RPARAN SQLPARAN IDENTIFIER SQRPARAN SEMICOLON
        """
        tag_name = token[7] if len(token) == 10 else None
        channels = []
        for name in token[2]:
            buffer_size = token[4]
            if (isinstance(buffer_size, str) or isinstance(buffer_size, unicode)) \
                and buffer_size == "*":
                buffer_size = -1
            channels.append(Channel(name, buffer_size, tag_name))
        return channels


class ModelParserExtension(LexYaccParserExtension):
    """
    Extension for parsing functions
    """
    
    def __init__(self):
        LexYaccParserExtension.__init__(self)
        
        self.builder = Builder()
        
    ##########################################
    #           RESERVED WORDS
    ##########################################
    
    def word_channels_specification(self, t):
        t.lexer.push_state('channels')
        return t
    
    ##########################################
    #                TOKENS
    ##########################################

    def token_block_close(self, t):
        r'}'
        
        t.lexer.pop_state()
        return t
    
    ##########################################
    #                RULES
    ##########################################
    
    def channels_specification(self, t):
        """
        specification : CHANNELS_SPECIFICATION BLOCKOPEN channels_list BLOCKCLOSE
        """
        pass
    
    
    def channels_list(self, t):
        """
        channels_list : channel 
                    | channels_list channel
        """
        pass
    
    def channel(self, t):
        """
        channel : CHANNEL identifiers_list LPARAN channel_buffor RPARAN SEMICOLON
            | CHANNEL identifiers_list LPARAN channel_buffor RPARAN SQLPARAN IDENTIFIER SQRPARAN SEMICOLON
        """
        for ch in self.builder.build_channels(t):
            self.parser.store.channels.append(ch)
    
    def channel_buffor(self, t):
        """
        channel_buffor : STAR
            | INTEGER
        """
        t[0] = t[1]
    
    def _extend(self):
        
        self.parser.add_state('channels', 'inclusive')

        self.parser.add_reserved_word('channels', 'CHANNELS_SPECIFICATION', func=self.word_channels_specification)
        self.parser.add_reserved_word('channel', 'CHANNEL', state='channels')

        self.parser.add_token('BLOCKCLOSE', func=self.token_block_close, states=['channels'])
        
        self.parser.add_rule(self.channels_specification)
        self.parser.add_rule(self.channels_list)
        self.parser.add_rule(self.channel)
        self.parser.add_rule(self.channel_buffor)
        

    


'''
Created on 31-05-2013

@author: Damian Rusinek <damian.rusinek@gmail.com>
'''


class TimeTrace():
    
    def __init__(self, host, process, instruction, expressions_details, started_at, length):
        """ """
        self.host = host
        self.process = process
        self.instruction = instruction
        self.expressions_details = expressions_details
        self.started_at = started_at  # time in ms
        self.length = length  # time in ms


class ChannelMessageTrace():
    
    def __init__(self, channel, message,
                 sender, sent_at, sending_time,
                 receiver, started_waiting_at,
                 started_receiving_at, receiving_time):
        """ """
        self.channel = channel
        self.message = message
        self.sender = sender
        self.sent_at = sent_at  # time in ms
        self.sending_time = sending_time  # time in ms
        self.receiver = receiver
        self.started_waiting_at = started_waiting_at  # time in ms
        self.started_receiving_at = started_receiving_at  # time in ms
        self.receiving_time = receiving_time  # time in ms
        



#!/bin/env/python

import copy
#from aqopa.simulator.error import RuntimeException

'''
Created on 23-04-2013

@author: Damian Rusinek <damian.rusinek@gmail.com>
'''

################################################
#             Names functions
################################################
        
def original_name(name):
    """
    Return name without indexes.
    """
    return name.split('.')[0]

def name_indexes(name):
    """
    Return indexes of name.
    """
    indexes = name.split('.')
    if len(indexes) == 0:
        indexes = []
    else:
        indexes = indexes[1:]
    return [ int(i) for i in indexes ]

################################################
#              QoPML Model Elements
################################################


class Function():
    
    def __init__(self, name, params = [], qop_params = [], comment = ""):
        self.name = name
        self.params = params
        self.qop_params = qop_params
        self.comment = comment
        
    def clone(self):
        return Function(copy.copy(self.name), copy.deepcopy(self.qop_params), copy.copy(self.comment))
        
    def __unicode__(self):
        
        comment = ""
        if self.comment:
            comment = "( %s )" % self.comment
            
        qop_params = ""
        if len(self.qop_params):
            qop_params = [ '%s : %s' % (name, ', '.join(pars)) for (name, pars) in self.qop_params ]
            qop_params = ', '.join(qop_params)
            qop_params = "[%s]" % qop_params
            
        return u"fun %s(%s) %s %s" % (self.name, ', '.join(self.params), qop_params, comment)
        

COMMUNICATION_TYPE_IN = 1
COMMUNICATION_TYPE_OUT = 2
    
class Channel():
    
    def __init__(self, name, buffor_size, tag_name=None):
        self.name = name
        self.buffor_size = buffor_size # Buffor size < 0 is unlimited
        self.tag_name = tag_name

    def __unicode__(self):
        buffor_size = str(self.buffor_size) if self.buffor_size >= 0 else "*"
        return u"channel %s [%s]" % (self.name, buffor_size)
    
    def clone(self):
        return Channel(copy.copy(self.name), copy.copy(self.buffor_size))

class Equation():
    
    def __init__(self, simple, composite):
        self.simple = simple
        self.composite = composite
        
    def __unicode__(self):
        return u"eq %s = %s" % ( unicode(self.composite), unicode(self.simple) )
    
    def clone(self):
        return Equation(self.simple.clone(), self.composite.clone())
    
################################
#        Expressions
################################
    
class BooleanExpression():
    
    def __init__(self, val):
        self.val = bool(val)
        
    def clone(self):
        return BooleanExpression(copy.copy(self.val))
        
    def __unicode__(self):
        return u"true" if self.val else u"false"
    
    def is_true(self):
        return self.val
    
class IdentifierExpression():
    
    def __init__(self, identifier):
        self.identifier = identifier
        
    def __unicode__(self):
        return unicode(self.identifier)
    
    def clone(self):
        return IdentifierExpression(copy.copy(self.identifier))
    
class CallFunctionExpression():
    
    def __init__(self, function_name, arguments=[], qop_arguments=[]):
        self.function_name = function_name
        self.arguments = arguments
        self.qop_arguments = qop_arguments
        
    def __unicode__(self):
        u = u"%s(%s)" % (unicode(self.function_name), unicode(', '.join([unicode(a) for a in self.arguments])))
        if len(self.qop_arguments) > 0:
            u += "[%s]" % unicode(', '.join([unicode(a) for a in self.qop_arguments]))
        return u

    def clone(self):
        # Regular clone
        expr = CallFunctionExpression(copy.copy(self.function_name),
                                      [a.clone() for a in self.arguments],
                                      [copy.copy(a) for a in self.qop_arguments])

        # Copy additional values (may come from predefined functions or keep calculated size, etc.)
        regular_vars = ['function_name', 'arguments', 'qop_arguments']
        for attr_name in self.__dict__:
            if attr_name not in regular_vars:
                setattr(expr, attr_name, copy.deepcopy(self.__dict__[attr_name]))
        return expr


COMPARISON_TYPE_EQUAL = 1
COMPARISON_TYPE_NOT_EQUAL = 2

class ComparisonExpression():
    
    def __init__(self, left, right, comparison_type):
        self.left = left
        self.right = right
        self.comparison_type = comparison_type
        
    def __unicode__(self):
        return "%s %s %s" % (unicode(self.left), u'==' if self.is_equal_type() else '!=',
                             unicode(self.right))

    def is_equal_type(self):
        return self.comparison_type == COMPARISON_TYPE_EQUAL
    
    def clone(self):
        return ComparisonExpression(self.left.clone(), self.right.clone(), self.comparison_type)
    
class TupleExpression():
    
    def __init__(self, elements):
        self.elements = elements
        
    def __unicode__(self):
        return u"(%s)" % unicode(', '.join([unicode(e) for e in self.elements]))
    
    def clone(self):
        return TupleExpression([e.clone() for e in self.elements])
    
class TupleElementExpression():
    
    def __init__(self, variable_name, index):
        self.variable_name = variable_name
        self.index = index
        
    def __unicode__(self):
        return u"%s[%d]" % (unicode(self.variable_name), self.index)
    
    def clone(self):
        return TupleElementExpression(copy.copy(self.variable_name),
                                      copy.copy(self.index))
      
################################
#        Instructions
################################
    
class CallFunctionInstruction():
    
    def __init__(self, function_name, arguments=[], qop_arguments=[]):
        self.function_name = function_name
        self.arguments = arguments
        self.qop_arguments = qop_arguments
        
    def __unicode__(self):
        return u"%s(%s)[%s];" % ( unicode(self.function_name), unicode(', '.join([ unicode(a) for a in self.arguments])),
                                  unicode(', '.join([ unicode(a) for a in self.qop_arguments])) )
    
    def clone(self):
        return CallFunctionInstruction(copy.copy(self.function_name), 
                                      [ a.clone() for a in self.arguments ],
                                      [ copy.copy(a) for a in self.qop_arguments ])
        
class FinishInstruction():
    
    def __init__(self, command):
        self.command = command
        
    def clone(self):
        return FinishInstruction(copy.copy(self.command))
        
    def __unicode__(self):
        return "%s;" % unicode(self.command)
    
class ContinueInstruction():
    
    def clone(self):
        return ContinueInstruction()
    
    def __unicode__(self):
        return u"continue;"

class BreakInstruction():

    def clone(self):
        return BreakInstruction()

    def __unicode__(self):
        return u"break;"
    
class AssignmentInstruction():
    
    def __init__(self, variable_name, expression):
        self.variable_name = variable_name
        self.expression = expression
        
    def clone(self):
        return AssignmentInstruction(copy.copy(self.variable_name), 
                                     self.expression.clone())
        
    def __unicode__(self):
        return u"%s = %s;" % (unicode(self.variable_name), unicode(self.expression))


class CommunicationInstruction():
    
    def __init__(self, communication_type, channel_name, variable_name, filters):
        self.communication_type = communication_type
        self.channel_name = channel_name
        self.variable_name = variable_name
        self.filters = filters
        
    def clone(self):
        filters = []
        for f in self.filters:
            if isinstance(f, basestring):
                filters.append(copy.copy(f))
            else:
                filters.append(f.clone())
        return CommunicationInstruction(copy.copy(self.communication_type),
                                        copy.copy(self.channel_name), 
                                        copy.copy(self.variable_name),
                                        filters)
        
    def is_out(self):
        return self.communication_type == COMMUNICATION_TYPE_OUT
        
    def __unicode__(self):
        if self.communication_type == COMMUNICATION_TYPE_IN:
            filters_str = u""
            if len(self.filters) > 0:
                filters_str = u", ".join([unicode(f) for f in self.filters])
                filters_str = u": |%s|" % filters_str
            return u"in (%s: %s%s);" % (unicode(self.channel_name),
                                        unicode(self.variable_name),
                                        filters_str)
        else:
            return u"out (%s: %s);" % (unicode(self.channel_name), unicode(self.variable_name))


class IfInstruction():
    
    def __init__(self, condition, true_instructions, false_instructions=[]):
        self.condition = condition
        self.true_instructions = true_instructions
        self.false_instructions = false_instructions
        
    def clone(self):
        t_instructions = []
        for i in self.true_instructions:
            t_instructions.append(i.clone())
        f_instructions = []
        for i in self.false_instructions:
            f_instructions.append(i.clone())
        return IfInstruction(self.condition.clone(), t_instructions, f_instructions)
        
    def __unicode__(self):
        return u"if (%s) ..." % unicode(self.condition)
        
class WhileInstruction():
    
    def __init__(self, condition, instructions):
        self.condition = condition
        self.instructions = instructions
        
    def __unicode__(self):
        return u"while (%s) ..." % unicode(self.condition)

    def clone(self):
        instructions = []
        for i in self.instructions:
            instructions.append(i.clone())
        return WhileInstruction(self.condition.clone(), instructions)
        
################################
#        Hosts
################################

class HostSubprocess():
    
    def __init__(self, name, instructions_list):
        self.name = name
        self.instructions_list = instructions_list
        self.all_channels_active = False
        self.active_channels = []
        
    def clone(self):
        
        instructions_list = []
        for i in self.instructions_list:
            instructions_list.append(i.clone())
        
        p = HostSubprocess(copy.copy(self.name), instructions_list)
        p.all_channels_active = copy.copy(self.all_channels_active)
        p.active_channels = copy.deepcopy(self.active_channels)
        return p
        
    def __unicode__(self):
        return u"subprocess %s" % unicode(self.name)

class HostProcess():
    
    def __init__(self, name, instructions_list):
        self.name = name
        self.instructions_list = instructions_list
        self.all_channels_active = False
        self.active_channels = []
        
        self.follower = None
        
    def clone(self):
        
        instructions_list = []
        for i in self.instructions_list:
            instructions_list.append(i.clone())
        
        p = HostProcess(copy.copy(self.name), instructions_list)
        p.all_channels_active = copy.copy(self.all_channels_active)
        p.active_channels = copy.deepcopy(self.active_channels)
        return p
        
    def __unicode__(self):
        return u"process %s" % unicode(self.name)
        
        
class Host():
    
    def __init__(self, name, schedule_algorithm, instructions_list, predefined_values=[]):
        self.name = name
        self.schedule_algorithm = schedule_algorithm
        self.instructions_list = instructions_list
        self.predefined_values = predefined_values
        self.all_channels_active = False
        self.active_channels = []
        
    def __unicode__(self):
        return u"host %s" % unicode(self.name)
        
    def clone(self):
        
        instructions_list = []
        for i in self.instructions_list:
            if isinstance(i, HostProcess):
                instructions_list.append(i.clone())
            else:
                instructions_list.append(i)
        
        predefined_values = []
        for i in self.predefined_values:
            predefined_values.append(i.clone())
        
        h = Host(copy.copy(self.name), copy.copy(self.schedule_algorithm), instructions_list, predefined_values)
        h.all_channels_active = self.all_channels_active
        h.active_channels = self.active_channels
        return h
        
################################
#        Versions
################################

class Version():
    
    def __init__(self, name):
        self.name = name
        self.run_hosts = []
        self.metrics_sets = []
        self.communication = {'mediums': {}}
        
    def __unicode__(self):
        return u"version %d" % self.name
    
    def clone(self):
        r_hosts = []
        for rh in self.run_hosts:
            r_hosts.append(rh.clone())
        m_sets = []
        for ms in self.metrics_sets:
            m_sets.append(ms.clone())
        v = Version(copy.copy(self.name))
        v.run_hosts = r_hosts
        v.metrics_sets = m_sets
        return v
    
class VersionRunHost():
    
    def __init__(self, host_name):
        self.host_name = host_name
        self.all_channels_active = False
        self.active_channels = []
        self.repetitions = 1
        self.repeated_channels = [] # deprecated
        self.run_processes = []
        
    def __unicode__(self):
        return u"run host %s (...)" % self.host_name
    
    def clone(self):
        r_processes = []
        for rp in self.run_processes:
            r_processes.append(rp.clone())
        rh = VersionRunHost(copy.copy(self.host_name))
        rh.all_channels_active = self.all_channels_active
        rh.active_channels = copy.deepcopy(self.active_channels)
        rh.repetitions = self.repetitions
        rh.repeated_channels = copy.deepcopy(self.repeated_channels)
        rh.run_processes = r_processes
        return rh
    
class VersionRunProcess():
    
    def __init__(self, process_name):
        self.process_name = process_name
        self.all_subprocesses_active = False
        self.active_subprocesses = []
        self.repetitions = 1
        self.repeated_channels = [] # deprecated
        self.follower = None
        
    def __unicode__(self):
        s = u"run %s (...)" % self.process_name
        if self.follower:
            s += " -> %s" % unicode(self.follower)
        return s
    
    def clone(self):
        rp = VersionRunProcess(copy.copy(self.process_name))
        rp.all_subprocesses_active = self.all_subprocesses_active
        rp.active_subprocesses = copy.deepcopy(self.active_subprocesses)
        rp.repetitions = self.repetitions
        rp.repeated_channels = copy.deepcopy(self.repeated_channels)
        if self.follower:
            rp.follower = self.follower.clone()
        return rp

################################
#        Metrics
################################

class MetricsConfiguration():
    
    def __init__(self, name, specifications):
        self.name = name
        self.specifications = specifications
        
    def __unicode__(self):
        return u"conf %s { ... }" % self.name
    
    def clone(self):
        return MetricsConfiguration(copy.copy(self.name), copy.deepcopy(self.specifications))

class MetricsSet():
    
    def __init__(self, host_name, configuration_name):
        self.host_name = host_name
        self.configuration_name = configuration_name
    
    def __unicode__(self):
        return u"set host %s (%s) { ... }" % (self.host_name, self.configuration_name)
    
    def clone(self):
        return MetricsSet(copy.copy(self.host_name), copy.copy(self.configuration_name))
        
class MetricsData():
    
    def __init__(self, name, blocks, plus = False, star = False):
        self.name = name
        self.blocks = blocks
        self.star = star
        self.plus = plus
        
    def __unicode__(self):
        return u"data %s { ... }" % self.name
        
    def clone(self):
        blocks = []
        for b in self.blocks:
            blocks.append(b.clone())
        return MetricsData(copy.copy(self.name), blocks, copy.copy(self.plus), copy.copy(self.star))
        
class MetricsPrimitiveBlock():
    
    def __init__(self, header, metrics):
        self.header = header
        self.metrics = metrics
        
    def clone(self):
        metrics = []
        for m in self.metrics:
            metrics.append(m.clone())
        return MetricsPrimitiveBlock(self.header.clone(), metrics)
        
class MetricsPrimitiveHeader():
    
    def __init__(self, params, services_params):
        self.params = params
        self.services_params = services_params
        
    def clone(self):
        params = []
        for p in self.params:
            params.append(p.clone())
        s_params = []
        for p in self.services_params:
            s_params.append(p.clone())
        return MetricsPrimitiveHeader(params, s_params)
    
class MetricsServiceParam():
    
    def __init__(self, service_name, param_name, unit=None):
        self.service_name = service_name
        self.param_name = param_name
        self.unit = unit
        
    def clone(self):
        return MetricsServiceParam(copy.copy(self.service_name),
                                   copy.copy(self.param_name),
                                   copy.copy(self.unit))
        
class MetricsPrimitive():
    
    def __init__(self, arguments):
        self.arguments = arguments
        
    def clone(self):
        return MetricsPrimitive(copy.deepcopy(self.arguments))
      
################################
#        Communication
################################

class TopologyRuleHost():
    
    def __init__(self, identifier, index_range=None, i_shift=None):
        self.identifier = identifier
        self.index_range = index_range
        self.i_shift = i_shift

    def __unicode__(self):
        s = self.identifier
        if self.index_range is not None:
            s += "["
            if self.index_range[0] is not None:
                s += str(self.index_range[0])
            s += ":"
            if self.index_range[1] is not None:
                s += str(self.index_range[1])
            s += "]"
        elif self.i_shift is not None:
            s += "[i"
            if self.i_shift >= 0:
                s += "+"
            s += str(self.i_shift) + "]"
        return unicode(s)

class TopologyRule():
    
    def __init__(self, left_host, arrow, right_host, parameters=None):
        self.left_host = left_host
        self.right_host = right_host
        self.arrow = arrow
        self.parameters = parameters if parameters is not None else {}

    def __unicode__(self):
        params_list = []
        for p in self.parameters:
            params_list.append(u"{0}={1}".format(p, self.parameters[p]))
        params_str = u""
        if len(params_list) > 0:
            params_str = u" : " + u", ".join(params_list)
        right_host_str = u"*"
        if self.right_host is not None:
            right_host_str = unicode(self.right_host)
        return u"{0} {1} {2}{3};".format(unicode(self.left_host), self.arrow, right_host_str, params_str)
      
################################
#       Algorithms
################################

class AlgCallFunction():
    
    def __init__(self, function_name, args):
        self.function_name = function_name
        self.args = args
        
class AlgWhile():
    
    def __init__(self, condition, instructions):
        self.condition = condition
        self.instructions = instructions
        
class AlgIf():
    
    def __init__(self, condition, true_instructions, false_instructions):
        self.condition = condition
        self.true_instructions = true_instructions
        self.false_instructions = false_instructions
        
class AlgReturn():
    
    def __init__(self, expression):
        self.expression = expression
        
class AlgAssignment():
    
    def __init__(self, identifier, expression):
        self.identifier = identifier
        self.expression = expression
    


'''
Created on 05-12-2013

@author: Damian Rusinek <damian.rusinek@gmail.com>
'''
import sys

from aqopa.simulator.state import Hook

class PrintResultsHook(Hook):

    def __init__(self, module, simulator, output_file=sys.stdout):
        """ """
        self.module = module
        self.simulator = simulator
        self.output_file = output_file
        
    def execute(self, context, **kwargs):
        """ """

        self.output_file.write('-'*80)
        self.output_file.write('\n')
        self.output_file.write('Module\tEnergy Analysis (consumption in J and Ah)')
        self.output_file.write('\n')
        self.output_file.write('Version\t%s\n' % self.simulator.context.version.name)
        self.output_file.write('\n')

        voltage = 3.0
        self.module.set_voltage(voltage)

        consumptions = self.module.get_hosts_consumptions(self.simulator, context.hosts, voltage)
        for h in context.hosts:
            host_consumptions = {
                'energy': 'N/A',
                'amp-hour': 'N/A',
            }
            if h in consumptions:
                host_consumptions = consumptions[h]
            self.output_file.write('{0}\t{1}\t{2}\t'.format(h.name, str(host_consumptions['energy']),
                                                            str(host_consumptions['amp-hour'])))
            self.output_file.write("\n")
        self.output_file.write('\n')



'''
Created on 03-04-2014

@author: damian
'''
from aqopa.model import AlgWhile, TupleExpression, AlgCallFunction, AlgReturn, AlgIf, AlgAssignment
from aqopa.simulator.error import RuntimeException


class AlgorithmResolver():

    def __init__(self):
        self.algorithms = {}

    def add_algorithm(self, name, algorithm):
        self.algorithms[name] = algorithm

    def has_algorithm(self, name):
        return name in self.algorithms

    def get_algorithm(self, name):
        return self.algorithms[name]

    def calculate(self, context, host, alg_name, variables=None):
        if not self.has_algorithm(alg_name):
            return 0
        if variables is None:
            variables = {}
        return AlgorithmCalculator(context, host, alg_name, variables, self.algorithms[alg_name]).calculate()


class AlgorithmCalculator():

    def __init__(self, context, host, algorithm_name, variables, algorithm):
        self.context = context
        self.host = host
        self.algorithm_name = algorithm_name
        self.variables = variables
        self.algorithm = algorithm
        self.instructions = algorithm['instructions']
        self.link_quality = variables['link_quality'] if 'link_quality' in variables else 1

        self.instructions_stack = [algorithm['instructions']]
        self.instructions_stack_index = 0
        self.instructions_lists_indexes = {0: 0}

        self.left_associative_operators = ['--', '-', '+', '*', '/', '==', '!=', '<=', '>=', '>', '<', '&&', '||']
        self.right_associative_operators = []
        self.all_operators = self.left_associative_operators + self.right_associative_operators

        self.return_value = None

    def get_index_in_current_list(self):
        """ Returns the index of instruction in current list """
        return self.instructions_lists_indexes[self.instructions_stack_index]

    def in_main_stack(self):
        """ Returns True when current instruction is in main stack """
        return self.instructions_stack_index == 0

    def has_current_instruction(self):
        """ """
        return self.get_index_in_current_list() < len(self.instructions_stack[self.instructions_stack_index])

    def get_current_instruction(self):
        """ Returns the instruction that should be executed next """
        return self.instructions_stack[self.instructions_stack_index][self.get_index_in_current_list()]

    def finished(self):
        """ Returns True when algorithm is finished """
        return self.return_value is not None or (not self.has_current_instruction() and self.in_main_stack())

    def goto_next_instruction(self):
        """ """
        self.instructions_lists_indexes[self.instructions_stack_index] += 1
        while not self.finished() and not self.has_current_instruction():
            self.instructions_stack.pop()
            del self.instructions_lists_indexes[self.instructions_stack_index]
            self.instructions_stack_index -= 1

            if not self.finished():
                if not isinstance(self.get_current_instruction(), AlgWhile):
                    self.instructions_lists_indexes[self.instructions_stack_index] += 1

    def add_instructions_list(self, instructions):
        """ Adds new instructions list to the stack """
        self.instructions_stack.append(instructions)
        self.instructions_stack_index += 1
        self.instructions_lists_indexes[self.instructions_stack_index] = 0

    def calculate_function_value(self, call_function_instruction):
        if call_function_instruction.function_name == 'size':
            var_name = call_function_instruction.args[0]
            if var_name not in self.variables:
                raise RuntimeException("Variable {0} not defined in communication algorithm {1}."
                                       .format(var_name, self.algorithm_name))
            value = self.variables[var_name]
            # If tuple element expression
            if len(call_function_instruction.args) > 1:
                if not isinstance(value, TupleExpression):
                    raise RuntimeException("Variable {0} in communication algorithm {1} is not tuple, it is: {2}."
                                           .format(var_name, self.algorithm_name, unicode(value)))
                index = call_function_instruction.args[1]
                if len(value.elements) <= index:
                    raise RuntimeException("Variable {0} in communication algorithm {1} has "
                                           "{2} elements while index {3} is asked."
                                           .format(var_name, self.algorithm_name, len(value.elements), index))
                value = value.elements[index]
            return self.context.metrics_manager.get_expression_size(value, self.context, self.host)
        elif call_function_instruction.function_name == 'quality':
            if self.link_quality is None:
                raise RuntimeException("Link quality is undefined in {0} algorithm. "
                                       .format(self.algorithm_name))
            return self.link_quality
        raise RuntimeException("Unresolved reference to function {0}() in algorithm {1}."
                               .format(call_function_instruction.function_name, self.algorithm_name))

    def _is_operation_token(self, token):
        return isinstance(token, basestring) and token in self.all_operators

    def _operator_order(self, operator):
        """
        Returns the order of operator as number.
        """
        orders = [['==', '!=', '<=', '>=', '>', '<', '&&', '||'], ['--', '-', '+'], ['*', '/']]
        for i in range(0, len(orders)):
            if operator in orders[i]:
                return i
        raise RuntimeException("Operator {0} undefined in algorithm {1}.".format(operator, self.algorithm_name))

    def _make_rpn(self, expression):
        """ """
        stack = []
        rpn = []
        for token in expression:
            # if operator
            if self._is_operation_token(token):
                while len(stack) > 0:
                    top_operator = stack[len(stack)-1]
                    # if current operator is left-associative and its order is lower or equal than top operator
                    # or current operator is right-associative and its order is lower than top operator
                    if (token in self.left_associative_operators
                            and self._operator_order(token) <= self._operator_order(top_operator))\
                       or (token in self.right_associative_operators
                            and self._operator_order(token) < self._operator_order(top_operator)):
                        rpn.append(stack.pop())
                    else:
                        break
                stack.append(token)
            elif token == '(':
                stack.append(token)
            elif token == ')':
                found_paran = False
                while len(stack) > 0:
                    top_operator = stack[len(stack)-1]
                    if top_operator == '(':
                        found_paran = True
                        stack.pop()
                        break
                    else:
                        rpn.append(stack.pop())
                if not found_paran:
                    raise RuntimeException("Incorrect number of brackets in algorithm {0}.".format(self.algorithm_name))
            else:  # else number
                if isinstance(token, AlgCallFunction):
                    token = self.calculate_function_value(token)
                elif isinstance(token, basestring):
                    if token not in self.variables:
                        raise RuntimeException("Variable {0} not defined in communication algorithm {1}."
                                               .format(token, self.algorithm_name))
                    token = self.variables[token]
                rpn.append(float(token))
        while len(stack) > 0:
            rpn.append(stack.pop())
        return rpn

    def _calculate_operation(self, operator, left, right):
        """ """
        if operator == '+':
            return left + right
        elif operator == '-':
            return left - right
        elif operator == '*':
            return left * right
        elif operator == '/':
            return left / right
        elif operator == '==':
            return left == right
        elif operator == '!=':
            return left != right
        elif operator == '>=':
            return left >= right
        elif operator == '>':
            return left > right
        elif operator == '<=':
            return left <= right
        elif operator == '<':
            return left < right
        else:
            raise RuntimeException("Incorrect operator {0} of brackets in algorithm {1}."
                                   .format(operator, self.algorithm_name))

    def _calculate_rpn(self, rpn_elements):
        """ """
        stack = []
        for token in rpn_elements:
            # if operator
            if self._is_operation_token(token):
                if token == '--':
                    value = stack.pop()
                    value = - value
                    stack.append(value)
                else:
                    a = stack.pop()
                    b = stack.pop()
                    stack.append(self._calculate_operation(token, b, a))
            else:  # number
                stack.append(token)
        return stack.pop()

    def calculate_value(self, expression):
        rpn_elements = self._make_rpn(expression)
        return self._calculate_rpn(rpn_elements)

    def execute_current_instruction(self):
        current_instruction = self.get_current_instruction()
        if isinstance(current_instruction, AlgReturn):
            self.return_value = self.calculate_value(current_instruction.expression)
            self.goto_next_instruction()
        elif isinstance(current_instruction, AlgWhile):
            if len(current_instruction.instructions) > 0 and self.calculate_value(current_instruction.condition):
                self.add_instructions_list(current_instruction.instructions)
            else:
                self.goto_next_instruction()
        elif isinstance(current_instruction, AlgIf):
            if self.calculate_value(current_instruction.condition):
                instructions = current_instruction.true_instructions
            else:
                instructions = current_instruction.false_instructions
            if len(instructions) > 0:
                self.add_instructions_list(instructions)
            else:
                self.goto_next_instruction()
        elif isinstance(current_instruction, AlgAssignment):
            self.variables[current_instruction.identifier] = self.calculate_value(current_instruction.expression)
            self.goto_next_instruction()

    def calculate(self):
        while not self.finished():
            self.execute_current_instruction()
        if self.return_value is None:
            raise RuntimeException("Algorithm {0} has no return value. Did you forget to use return instruction?"
                                   .format(self.algorithm_name))
        return self.return_value


'''
Created on 27-12-2013

@author: Damian Rusinek <damian.rusinek@gmail.com>
'''
from aqopa.model import IfInstruction, WhileInstruction, AssignmentInstruction
from aqopa.model.parser.lex_yacc import LexYaccParser, LexYaccParserExtension
from aqopa.simulator.state import Process


class ComputionalParserExtension(LexYaccParserExtension):

    def __init__(self):
        LexYaccParserExtension.__init__(self)

        self.open_blocks_cnt = 0

    ##########################################
    #                TOKENS
    ##########################################

    def token_identifier(self, t):
        r'[_a-zA-Z][_a-zA-Z0-9]*'

        words = self.parser.get_reserved_words()

        states_stack = []
        states_stack.extend(t.lexer.lexstatestack)
        states_stack.append(t.lexer.current_state())

        i = len(states_stack)-1
        while i >= 0:
            state = states_stack[i]
            if state in words:
                state_words = words[state]
                for state_word in state_words:

                    tvalue = t.value
                    state_word_value = state_word
                    word_tuple = state_words[state_word]

                    # if not case sensitive
                    if not word_tuple[2]:
                        tvalue = tvalue.lower()
                        state_word_value = state_word_value.lower()

                    if tvalue == state_word_value:
                        # If function exists
                        if word_tuple[1]:
                            t = word_tuple[1](t)
                        t.type = word_tuple[0]
                        break
            i -= 1

        return t

    def token_float(self, t):
        r'([1-9][0-9]*\.[0-9]+)|(0\.[0-9]+)'
        t.value = float(t.value)
        return t

    def token_integer(self, t):
        r'0|[1-9][0-9]*'
        t.value = int(t.value)
        return t

    def token_blockopen(self, t):
        r"\{"
        self.open_blocks_cnt += 1
        return t

    def token_blockclose(self, t):
        r"\}"
        self.open_blocks_cnt -= 1
        if self.open_blocks_cnt == 0:
            t.lexer.pop_state()
        return t

    ##########################################
    #                RULES
    ##########################################

    def rule_number(self, t):
        """
        number : FLOAT
                | INTEGER
        """
        t[0] = t[1]

    def reputation_expression_operations(self, t):
        """
        reputation_expression : reputation_expression PLUS reputation_expression
                            | reputation_expression MINUS reputation_expression
                            | reputation_expression TIMES reputation_expression
                            | reputation_expression DIVIDE reputation_expression
        """
        if t[2] == '+':
            t[0] = t[1] + t[3]
        elif t[2] == '-':
            t[0] = t[1] - t[3]
        elif t[2] == '/':
            t[0] = t[1] / t[3]
        elif t[2] == '*':
            t[0] = t[1] * t[3]

    def reputation_expression_values(self, t):
        """
        reputation_expression : number
                            | IDENTIFIER
        """
        try:
            t[0] = float(t[1])
        except ValueError:
            t[0] = self.parser.vars[t[1]]

    def reputation_expression_uminus(self, t):
        """
        reputation_expression : MINUS reputation_expression %prec UMINUS
        """
        t[0] = - t[2]

    def reputation_expression_paran(self, t):
        """
        reputation_expression : LPARAN reputation_expression RPARAN
        """
        t[0] = t[2]

    def reputation_expression_conditional_comparison(self, t):
        """
        reputation_expression_conditional : reputation_expression EQUAL EQUAL reputation_expression
                                        | reputation_expression EXCLAMATION EQUAL reputation_expression
                                        | reputation_expression GREATER reputation_expression
                                        | reputation_expression GREATER EQUAL reputation_expression
                                        | reputation_expression SMALLER reputation_expression
                                        | reputation_expression SMALLER EQUAL reputation_expression
                                        | reputation_expression AND AND reputation_expression
                                        | reputation_expression OR OR reputation_expression
        """
        if t[3] == '=':
            if t[2] == '=':
                t[0] = t[1] == t[4]
            elif t[2] == '!':
                t[0] = t[1] != t[4]
            elif t[2] == '>':
                t[0] = t[1] >= t[4]
            elif t[2] == '<':
                t[0] = t[1] <= t[4]
        elif t[2] == '>':
            t[0] = t[1] > t[3]
        elif t[2] == '<':
            t[0] = t[1] < t[3]
        elif t[2] == '&' and t[3] == '&':
            t[0] = t[1] and t[4]
        elif t[2] == '|' and t[3] == '|':
            t[0] = t[1] or t[4]
        else:
            t[0] = False

    def reputation_expression_conditional_operators(self, t):
        """
        reputation_expression_conditional : USED LPARAN IDENTIFIER RPARAN
                                        | EXISTS LPARAN IDENTIFIER RPARAN
        """
        if t[1].lower() == 'used':
            t[0] = self.parser.process_uses_variable(t[3])
        else:
            t[0] = self.parser.host.has_variable(t[3])


    def reputation_expression_conditional_paran(self, t):
        """
        reputation_expression_conditional : LPARAN reputation_expression_conditional RPARAN
        """
        t[0] = t[2]

    def _extend(self):
        """ """
        self.parser.add_reserved_word('if', 'IF', case_sensitive=True)
        self.parser.add_reserved_word('else', 'ELSE', case_sensitive=True)
        self.parser.add_reserved_word('used', 'USED', case_sensitive=True)
        self.parser.add_reserved_word('exists', 'EXISTS', case_sensitive=True)

        self.parser.add_token('FLOAT', func=self.token_float)
        self.parser.add_token('INTEGER', func=self.token_integer)
        self.parser.add_token('IDENTIFIER', func=self.token_identifier)

        self.parser.add_token('SEMICOLON', r';')
        self.parser.add_token('EQUAL', r'\=')

        self.parser.add_token('LPARAN', r'\(')
        self.parser.add_token('RPARAN', r'\)')
        self.parser.add_token('SQLPARAN', r'\[')
        self.parser.add_token('SQRPARAN', r'\]')
        self.parser.add_token('BLOCKOPEN', func=self.token_blockopen)
        self.parser.add_token('BLOCKCLOSE', func=self.token_blockclose)

        self.parser.add_token('PLUS', r'\+')
        self.parser.add_token('MINUS', r'\-')
        self.parser.add_token('TIMES', r'\*')
        self.parser.add_token('DIVIDE', r'/')

        self.parser.add_token('GREATER', r'\>')
        self.parser.add_token('SMALLER', r'\<')
        self.parser.add_token('EXCLAMATION', r'\!')
        self.parser.add_token('AND', r'\&')
        self.parser.add_token('OR', r'\|')

        self.parser.add_precedence(['PLUS', 'MINUS'], 'left')
        self.parser.add_precedence(['TIMES', 'DIVIDE'], 'left')
        self.parser.add_precedence(['UMINUS'], 'right')

        # Rules

        self.parser.add_rule(self.rule_number)
        self.parser.add_rule(self.reputation_expression_operations)
        self.parser.add_rule(self.reputation_expression_paran)
        self.parser.add_rule(self.reputation_expression_uminus)
        self.parser.add_rule(self.reputation_expression_values)
        self.parser.add_rule(self.reputation_expression_conditional_comparison)
        self.parser.add_rule(self.reputation_expression_conditional_operators)
        self.parser.add_rule(self.reputation_expression_conditional_paran)

class ComputionalParser(LexYaccParser):

    _conditional_expression_instance = None
    _simple_expression_instance = None

    def __init__(self):
        self.host = None
        self.vars = {}
        LexYaccParser.__init__(self)

    def restart(self):
        self.yaccer.restart()

    def parse_expression(self, s, host, rep_vars):
        self.host = host
        self.vars = rep_vars
        return self.yaccer.parse(input=s, lexer=self.lexer)

    def process_uses_variable(self, variable_name):
        """ Returns True if variable with given name is defined anywhere in process """
        index = 0
        instructions = self.host.get_current_process().instructions_list
        while index < len(instructions):
            instruction = instructions[index]
            if isinstance(instruction, IfInstruction):
                instructions.extend(instruction.true_instructions)
                instructions.extend(instruction.false_instructions)
            elif isinstance(instruction, WhileInstruction):
                instructions.extend(instruction.instructions)
            elif isinstance(instruction, AssignmentInstruction):
                if instruction.variable_name == variable_name:
                    return True
            index += 1
        return False

    @staticmethod
    def conditional_expression_instance():
        if ComputionalParser._conditional_expression_instance is None:
            parser_ext = ComputionalParserExtension()
            ComputionalParser._conditional_expression_instance = ComputionalParser()
            ComputionalParser._conditional_expression_instance.add_extension(parser_ext)
            ComputionalParser._conditional_expression_instance.start_symbol = 'reputation_expression_conditional'
            ComputionalParser._conditional_expression_instance.build()
        else:
            ComputionalParser._conditional_expression_instance.restart()

        return ComputionalParser._conditional_expression_instance

    @staticmethod
    def simple_expression_instance():
        if ComputionalParser._simple_expression_instance is None:
            parser_ext = ComputionalParserExtension()
            ComputionalParser._simple_expression_instance = ComputionalParser()
            ComputionalParser._simple_expression_instance.add_extension(parser_ext)
            ComputionalParser._simple_expression_instance.start_symbol = 'reputation_expression'
            ComputionalParser._simple_expression_instance.build()
        else:
            ComputionalParser._conditional_expression_instance.restart()

        return ComputionalParser._simple_expression_instance



def compute_conditional_expression(host, expression, vars):
    """
    Computes the result of conditional expression.
    """
    parser = ComputionalParser.conditional_expression_instance()
    return parser.parse_expression(expression, host, vars)

def compute_simple_expression(host, expression, vars):
    """
    Computes the result of simple expression.
    """
    parser = ComputionalParser.simple_expression_instance()
    return parser.parse_expression(expression, host, vars)

def update_vars(host, instructions, vars):
    """
    Updae vars according to the instructions of algorithm.
    """
    index = 0
    while index < len(instructions):
        instruction = instructions[index]
        if instruction['type'] == 'assignment':
            vars[instruction['identifier']] = compute_simple_expression(host, instruction['expression'], vars)
        elif instruction['type'] == 'if':
            condition = compute_conditional_expression(host, instruction['condition'], vars)
            if condition:
                if_instructions = instruction['true_instructions']
            else:
                if_instructions = instruction['false_instructions']
            if_index = index + 1
            for i in if_instructions:
                instructions.insert(if_index, i)
                if_index += 1
        index += 1
    return vars



#!/usr/bin/env python

import wx

"""
@file       main_notebook_gui.py
@brief      GUI for the main notebook, where we attach our AQoPA tabs
@author     Damian Rusinek <damian.rusinek@gmail.com>
@date       created on 05-09-2013 by Damian Rusinek
@date       edited on 09-05-2014 by Katarzyna Mazur (visual improvements mainly)
"""

class ResultsPanel(wx.Panel):
    """ """
    def __init__(self, *args, **kwargs):
        wx.Panel.__init__(self, *args, **kwargs)

        self.selectedModules = []
        self.moduleResultPanel = {}

        self._BuildMainLayout()

    def _BuildMainLayout(self):

        mainSizer = wx.BoxSizer(wx.VERTICAL)

        self.modulesChooserBox = wx.StaticBox(self, label="Modules")
        self.modulesChooserBox.Hide()
        self.modulesChooserBoxSizer = wx.StaticBoxSizer(self.modulesChooserBox, wx.HORIZONTAL)

        # create combobox, empty at first
        self.modulesChooserComboBox = wx.ComboBox(self, style=wx.TE_READONLY)
        # clear combo, do not remember prev choices
        self.modulesChooserComboBox.Clear()
        # DO NOT select, selecting can mess up many things
        self.modulesChooserComboBox.SetSelection(-1)
        self.modulesChooserComboBox.SetValue("")
        self.modulesChooserComboBox.Hide()
        # bind selection event - if user selects the module,
        # AQoPA will show up some results panels
        self.modulesChooserComboBox.Bind(wx.EVT_COMBOBOX, self.OnModuleSelected)

        # create static text - simple information for the user whatz goin' on with GUI
        self.modulesInfo = wx.StaticText(self, label="Select module to see the analysis results.")
        self.modulesInfo.Hide()

        # add text near the combocheckbox
        self.modulesChooserBoxSizer.Add(self.modulesInfo, 0, wx.ALL | wx.EXPAND, 5)
        # add some horizontal space - empty label its simple and effective
        self.modulesChooserBoxSizer.Add(wx.StaticText(self), 1, wx.ALL | wx.EXPAND, 5)
        # add combocheck box to the panel that will show up after selecting
        # modules in the modules tab from the main window
        self.modulesChooserBoxSizer.Add(self.modulesChooserComboBox, 1, wx.ALL | wx.EXPAND, 5)

        self.resultsBox = wx.StaticBox(self, label="Results")
        self.resultsBox.Hide()
        self.resultsBoxSizer = wx.StaticBoxSizer(self.resultsBox, wx.VERTICAL)

        mainSizer.Add(self.modulesChooserBoxSizer, 0, wx.ALL | wx.EXPAND, 5)
        # add an empty static text - sth like a vertical spacer
        for i in range(0, 2) :
            mainSizer.Add(wx.StaticText(self), 0, 0, wx.ALL | wx.EXPAND, 5)
        mainSizer.Add(self.resultsBoxSizer, 1, wx.ALL | wx.EXPAND, 5)

        self.SetSizer(mainSizer)
        self.Layout()

    def _BuildModulesLayout(self):
        """
        @brief builds a main layout
        """

        for m in self.selectedModules:
            if m in self.moduleResultPanel:
                continue

            gui = m.get_gui()

            resultPanel = gui.get_results_panel(self)
            self.resultsBoxSizer.Add(resultPanel, 1, wx.ALL | wx.EXPAND, 5)
            self.moduleResultPanel[m] = resultPanel

            self.Layout()
            resultPanel.Hide()

        # get unselected modules, make a list out of them
        uncheckedModules = []
        [uncheckedModules.append(m) for m in self.moduleResultPanel if m not in self.selectedModules]

        # delete unchecked panels [modules actually]
        for m in uncheckedModules:
            self.moduleResultPanel[m].Destroy()
            del self.moduleResultPanel[m]

        self.Layout()

    def SetSelectedModules(self, modules):
        """ """
        self.selectedModules = modules

        if len(self.selectedModules) > 0:
            self.resultsBox.Show()
            self.modulesChooserBox.Show()
            # DO NOT select, selecting can mess up many things
            # instead, clear combo and its selection
            self.modulesChooserComboBox.Clear()
            self.modulesChooserComboBox.SetSelection(-1)
            self.modulesChooserComboBox.SetValue("...")
            # add selected modules to the combobox
            for m in self.selectedModules:
                self.modulesChooserComboBox.Append(m.get_gui().get_name())
            # refresh combo and show it
            self.modulesChooserComboBox.Refresh()
            self.modulesChooserComboBox.Show()
            self.modulesInfo.Show()
        else:
            self.resultsBox.Hide()
            # clear'n' hide combobox if we do not need
            # to see analysis results (0 modules were selected)
            self.modulesChooserComboBox.Clear()
            self.modulesChooserComboBox.Hide()
            self.modulesChooserBox.Hide()
            self.modulesInfo.Hide()

        self._BuildModulesLayout()

    def ClearResults(self):
        """ """
        for m in self.selectedModules:
            gui = m.get_gui()
            gui.on_parsed_model()

    def OnModuleSelected(self, event):
        """
        @brief  when user selects the module in combobox,
        appropriate result panel shows up, yay!
        """

        # get current selection
        selectedModule = self.modulesChooserComboBox.GetValue()
        # hide all panels
        for m in self.moduleResultPanel:
            self.moduleResultPanel[m].Hide()
        # get the module object from the selected modules list,
        # simply find it by name - if its the same as the selected
        # on combo - that's the module we're looking for - grab it
        for m in self.selectedModules :
            if m.get_gui().get_name() == selectedModule :
                currentModule = m
                break
        # show the above-found module
        self.moduleResultPanel[currentModule].Show()
        self.Layout()


'''
Created on 07-09-2013

@author: Damian Rusinek <damian.rusinek@gmail.com>
'''
import sys

from aqopa.simulator.state import Hook

class PrintResultsHook(Hook):

    def __init__(self, module, simulator, output_file=sys.stdout):
        """ """
        self.module = module
        self.simulator = simulator
        self.output_file = output_file
        
    def execute(self, context, **kwargs):
        """ """

        self.output_file.write('-'*80)
        self.output_file.write('\n')
        self.output_file.write('Module\tTime Analysis (time in s)')
        self.output_file.write('\n')
        self.output_file.write('Version\t%s\n\n' % self.simulator.context.version.name)
        
        if self.simulator.infinite_loop_occured():
            self.output_file.write('ERROR\tInfinite loop on {0} -> {1}\n'.format(
                                unicode(self.simulator.context.get_current_host()),
                                unicode(self.simulator.context.get_current_instruction())))
            self.output_file.write('\n')

        for h in context.hosts:
            self.output_file.write('{0}\t{1}\t'.format(h.name, str(self.module.get_current_time(self.simulator, h))))
            if h.finished():
                self.output_file.write('Finished')
                if h.get_finish_error():
                    self.output_file.write(' with error\t{0}'.format(h.get_finish_error()))
            else:
                self.output_file.write('NOT Finished\t{0}'.format(unicode(h.get_current_instructions_context()\
                                                                .get_current_instruction())))
            self.output_file.write("\n")
            
        self.output_file.write('\n')
        self.output_file.write('Dropped messages\n')
        i = 0
        for c in context.channels_manager.channels:
            if c.get_dropped_messages_nb() > 0:
                i += 1
                self.output_file.write('%s\t%d\n' % (c.name, c.get_dropped_messages_nb()))
        if i == 0:
            self.output_file.write('None\n')
        self.output_file.write('\n')



from copy import copy, deepcopy
from aqopa import module
from aqopa.module.reputation.hook import PreInstructionHook
from aqopa.simulator.state import HOOK_TYPE_SIMULATION_FINISHED, HOOK_TYPE_PRE_INSTRUCTION_EXECUTION
from aqopa.model import CallFunctionExpression
from .parser import MetricsParserExtension, ModelParserExtension
from .gui import ModuleGui
from .console import PrintResultsHook

class Module(module.Module):
    """
    """
    
    def __init__(self):
        """ """
        self.guis = {}  # Divided by simulators - the reason for dict

        self.init_vars = {}  # Divided by name - the reason for dict
        self.algorithms = {}  # Divided by name - the reason for dict

        self.reputation_vars = {}  # Divided by hosts - the reason for dict

    def get_gui(self):
        if not getattr(self, '__gui', None):
            setattr(self, '__gui', ModuleGui(self))
        return getattr(self, '__gui', None)
    
    def extend_metrics_parser(self, parser):
        """
        Overriden
        """
        parser.add_extension(MetricsParserExtension())
        return parser

    def extend_model_parser(self, parser):
        """
        Overriden
        """
        parser.add_extension(ModelParserExtension(self))
        return parser
    
    def _install(self, simulator):
        """
        """
        hook = PreInstructionHook(self, simulator)
        simulator.register_hook(HOOK_TYPE_PRE_INSTRUCTION_EXECUTION, hook)
        return simulator
    
    def install_console(self, simulator):
        """ Install module for console simulation """
        self._install(simulator)
        hook = PrintResultsHook(self, simulator)
        simulator.register_hook(HOOK_TYPE_SIMULATION_FINISHED, hook)
        return simulator
        
    def install_gui(self, simulator):
        """ Install module for gui simulation """
        self._install(simulator)
        return simulator

    def get_algorithm(self, name):
        """
        Returns algorithm by name.
        """
        return deepcopy(self.algorithms[name])

    def set_reputation_var(self, host, var, val):
        """ """
        if host not in self.reputation_vars:
            self.reputation_vars[host] = {}
        self.reputation_vars[host][var] = val

    def get_host_vars(self, host):
        """
        """
        if host not in self.reputation_vars:
            return {}
        return copy(self.reputation_vars[host])





'''
Created on 07-05-2013

@author: Damian Rusinek <damian.rusinek@gmail.com>
'''
import copy
from aqopa.model import CommunicationInstruction, TupleExpression, ComparisonExpression, COMPARISON_TYPE_EQUAL, \
    CallFunctionExpression, AlgWhile, AlgReturn, AlgIf, AlgAssignment, AlgCallFunction
from aqopa.simulator.error import RuntimeException


class ChannelMessage():
    """
    Message sent through channel.
    """
    
    def __init__(self, sender, expression, expression_checker):
        self.sender = sender  # The host that sent the expression
        self.expression = expression  # The sent expression
        self.expression_checker = expression_checker  # checker used to check if message passes given filters

        self.not_for_requests = []  # List of requests that this expression should not been used by

    def cancel_for_request(self, request):
        self.not_for_requests.append(request)

    def is_for_request(self, request):
        return request not in self.not_for_requests

    def is_used(self):
        return len(self.not_for_hosts)

    def pass_filters(self, filters):
        """
        Returns True if message passes all given filters
        """
        if len(filters) == 0:
            return True
        else:
            if not isinstance(self.expression, TupleExpression) \
                    or (len(self.expression.elements) < len(filters)):
                return False

        for i in range(0, len(filters)):
            f = filters[i]
            if isinstance(f, basestring):  # star - accept everything
                continue
            else:
                cmp_expr = ComparisonExpression(f, self.expression.elements[i], COMPARISON_TYPE_EQUAL)
                if not self.expression_checker.result(cmp_expr, self.sender):
                    return False
        return True


class ChannelMessageRequest():
    """
    Representation of hosts' requests for messages.
    """
    
    def __init__(self, receiver, communication_instruction, expression_populator):
        """ """
        self.is_waiting = True
        self.receiver = receiver
        self.instruction = communication_instruction
        self.expression_populator = expression_populator  # populator used to populate current version of filters
        self.assigned_message = None

    def get_populated_filters(self):
        """
        Get list of filters with populated expressions - ready for comparison.
        """
        filters = []
        for f in self.instruction.filters:
            if f == '*':
                filters.append(f)
            else:
                filters.append(self.expression_populator.populate(f, self.receiver))
        return filters

    def clean(self):
        self.is_waiting = False
        self.assigned_message = None

    def start_waiting(self):
        self.is_waiting = True

    def get_message_from_buffer(self, buffer):
        """
        Adds messages from the given list and returns number of added messages.
        """
        # Get messages only if receiver is actually waiting
        filters = self.get_populated_filters()
        if self.is_waiting:
            if self.assigned_message is None and len(buffer) > 0:
                for message in buffer:
                    # Check if message passes the filters
                    if not message.pass_filters(filters):
                        continue
                    self.assigned_message = message
                    buffer.remove(message)
                    return True
        return False

    def ready_to_fulfill(self):
        """
        Return True when request has as many messages as he is waiting for.
        """
        return self.assigned_message is not None

    def fulfill(self):
        """
        Assigns requested variables with received messages.
        """
        if not self.ready_to_fulfill():
            raise RuntimeException("Request of host {0} in instruction {1} is not ready to fulfill."
                                   .format(self.receiver.name, unicode(self.instruction)))

        # Set variable sent in message
        self.receiver.set_variable(self.instruction.variable_name, self.assigned_message.expression.clone())

        # Move instructions context to the next instruction
        # print unicode(self.receiver.get_instructions_context_of_instruction(self.instruction).get_current_instruction())
        self.receiver.get_instructions_context_of_instruction(self.instruction)\
            .goto_next_instruction()
        # print unicode(self.receiver.get_instructions_context_of_instruction(self.instruction))
        self.receiver.mark_changed()


class Channel():
    """
    Simulation channel.
    """
    
    def __init__(self, name, buffer_size, tag_name=None):
        self.name = name
        self.tag_name = tag_name
        self._buffer_size = buffer_size   # Size of buffer, negative means that buffer is unlimited,
                                          # zero means that channel is asynhronous
        self._connected_hosts = []  # List of hosts that can use this channel
        self._connected_processes = []  # List of processes that can use this channel
        self._buffers = {}  # Hosts' buffers - the host is the key, the list of sent messages is the value

        self._waiting_requests = []
        self._dropped_messages_cnt = 0
        
    def connect_with_host(self, host):
        """
        Connect channel with host if it is not already connected.
        """
        if host in self._connected_hosts:
            return
        self._connected_hosts.append(host)

    def is_connected_with_host(self, host):
        """
        Returns True if host is connected with this channel.
        """
        return host in self._connected_hosts
        
    def connect_with_process(self, process):
        """
        Connect channel with process if it is not already connected.
        """
        if process in self._connected_processes:
            return
        self._connected_processes.append(process)
        
    def is_connected_with_process(self, process):
        """
        Returns True if process is connected with this channel.
        """
        return process in self._connected_processes

    def is_synchronous(self):
        """
        Returns True if channel is synchronous.
        """
        return self._buffer_size == 0
    
    def has_unlimited_buffer(self):
        """
        Returns True if channel has unlimited buffer.
        """
        return self._buffer_size < 0
    
    def get_dropped_messages_nb(self):
        """
        Return number of messages dropped on this channel
        """
        return self._dropped_messages_cnt

    def get_buffer_for_host(self, host):
        """
        Returns the content of buffer assigned to host.
        """
        if host not in self._buffers:
            self._buffers[host] = []
        return self._buffers[host]

    def get_existing_request_for_instruction(self, receiver, instruction):
        """
        Returns request on which channel is waiting for message or None if it is not waiting.
        """
        for i in range(0, len(self._waiting_requests)):
            request = self._waiting_requests[i]
            if request.instruction == instruction and request.receiver == receiver:
                return request
        return None

    def is_waiting_on_instruction(self, receiver, instruction):
        """
        Returns True if channel is waiting for message.
        """
        request = self.get_existing_request_for_instruction(receiver, instruction)
        if request:
            return request.is_waiting
        return False

    def wait_for_message(self, request):
        """
        Add host to the queue of hosts waiting for messages.
        Host can wait for many expressions.
        """
        if not self.is_connected_with_host(request.receiver):
            raise RuntimeException("Channel '%s' is not connected with host '%s'." % (self.name, request.receiver.name))

        # Check if this request already exists
        existing_request = self.get_existing_request_for_instruction(request.receiver, request.instruction)

        # If it does not exist add
        if not existing_request:
            # print 'IN', unicode(self.name), unicode(request.receiver), unicode(request.instruction)
            self._waiting_requests.append(request)
        else:
            # If request exists, check if it is waiting on IN instruction now
            if not existing_request.is_waiting:
                # If it is now (request stays in channel only to fill the buffer) start waiting on this request
                existing_request.start_waiting()

        # Check if request can be bound with expressions
        self._bind_messages_with_receivers()

    def get_filtered_requests(self, message, router):
        """
        Returns list of requests that can accept the message
        """
        requests = []
        for request in self._waiting_requests:
            # Check if there is a link between sender and receiver
            if not router.link_exists(self, message.sender, request.receiver):
                continue
            # Check if message has not been declined for waiting host
            if not message.is_for_request(request):
                continue
            # Check if message passes the filters
            filters = request.get_populated_filters()
            if not message.pass_filters(filters):
                continue
            requests.append(request)
        return requests

    def get_filtered_messages(self, request, router):
        """
        Returns list of messages from buffer that can be assigned to request
        """
        messages = []
        buffer = self.get_buffer_for_host(request.receiver)
        filters = request.get_populated_filters()
        for message in buffer:
            # Check if there is a link between sender and receiver
            if not router.link_exists(self, message.sender, request.receiver):
                continue
            # Check if message passes the filters
            if not message.pass_filters(filters):
                continue
            messages.append(message)
        return messages

    def send_message(self, sender_host, message, router):
        """
        Accept message with expressions.
        """
        if not self.is_connected_with_host(sender_host):
            raise RuntimeException("Channel '%s' is not connected with host '%s'." % (self.name, sender_host.name))

        # print 'OUT', unicode(self.name), unicode(sender_host), unicode(message.expression)

        # Put sent message in the buffers of receivers
        # Receivers are retrieved from the requests present in the channel
        # When the channel is synchronous the buffers are cleaned after the binding try
        # and requests are removed after they are fulfilled
        for request in self.get_filtered_requests(message, router):
            if request.receiver not in self._buffers:
                self._buffers[request.receiver] = []
            self._buffers[request.receiver].append(message)

        # Check if request can be bound with expressions
        self._bind_messages_with_receivers()

    def _bind_messages_with_receivers(self):
        """
        Assign messages from buffers to requests.
        """
        # Update all requests
        for request in self._waiting_requests:
            # Add messages from buffer to request
            # Add not more that request wants (it is one message)
            if request.receiver not in self._buffers:
                self._buffers[request.receiver] = []
            request.get_message_from_buffer(self._buffers[request.receiver])

            # Here the request is filled with all requested messages
            # or there was not enough messages in the buffer

            # If the request is filled with all requested messages = ready to fulfill
            if request.ready_to_fulfill():
                # Fulfill it
                request.fulfill()
                # print "Binded: ", unicode(request.instruction), unicode(request.receiver)

                # If channel is synchronous, delete the request - a new one will be created
                # when the intruction is executed again
                if self.is_synchronous():
                    self._waiting_requests.remove(request)
                else:
                    # If channel is asynchronous, clean the request - still accept messages to the buffer
                    request.clean()

        # Clean buffers if channel is synchronous
        if self.is_synchronous():
            for i in self._buffers:
                self._buffers[i] = []
        elif self._buffer_size > 0:  # Channel has a limit in buffer
            for i in self._buffers:
                if len(self._buffers[i]) > self._buffer_size:
                    self._dropped_messages_cnt += len(self._buffers[i]) - self._buffer_size
                    for j in range(self._buffer_size, len(self._buffers[i])):
                        del self._buffers[i][j]

    def add_left_messages_to_dropped(self):
        """
        If there are any undelivered messages, add their number to the number of dropped messages.
        """
        for i in self._buffers:
            self._dropped_messages_cnt += len(self._buffers[i])


class Manager():
    """ Channels manager class """
    
    def __init__(self, channels):
        self.channels = channels
        self.router = Router()

    def add_medium(self, name, topology, default_parameters):
        self.router.add_medium(name, topology, default_parameters)
        
    def has_medium(self, name):
        return self.router.has_medium(name)

    # def add_algorithm(self, name, algorithm):
    #     self.algorithm_resolver.add_algorithm(name, algorithm)
    #
    # def has_algorithm(self, name):
    #     return self.algorithm_resolver.has_algorithm(name)
    #
    # def get_algorithm(self, name):
    #     return self.algorithm_resolver.get_algorithm(name)

    def get_router(self):
        return self.router

    # def get_algorithm_resolver(self):
    #     return self.algorithm_resolver

    def find_channel(self, name, predicate=None):
        """
        """
        for ch in self.channels:
            if ch.name == name:
                if predicate is not None and not predicate(ch):
                    continue
                return ch
        return None
        
    def find_channel_for_current_instruction(self, context):
        """
        Finds channel connected with process/host from current instruction
        and with the channel name from instruction. 
        """
        instruction = context.get_current_instruction()
        if not isinstance(instruction, CommunicationInstruction):
            return None
        return self.find_channel_for_host_instruction(context, context.get_current_host(), 
                                                      instruction)
        
    def find_channel_for_host_instruction(self, context, host, instruction):
        """
        Finds channel connected with process/host from current instruction 
        of passed host and with the channel name from instruction. 
        """
        process = host.get_current_instructions_context().get_process_of_current_list()
        if process:
            return self.find_channel(instruction.channel_name,
                                     predicate=lambda x: x.is_connected_with_process(process))
        else:
            return self.find_channel(instruction.channel_name,
                                     predicate=lambda x: x.is_connected_with_host(host))

    def build_message(self, sender, expression, expression_checker):
        """
        Creates a message that will be sent.
        """
        return ChannelMessage(sender, expression, expression_checker)

    def build_message_request(self, receiver, communication_instruction,
                              expression_populator):
        """
        Creates a request for message when host executes instruction IN
        """
        return ChannelMessageRequest(receiver, communication_instruction, expression_populator)


class Router():
    
    def __init__(self):

        self.mediums = {}     # { MEDIUM_NAME -> {
                              #  'default_q' -> ...,
                              #  'default_...' -> ...,
                              #  'default_...' -> ...,

                              #  'topology' ->
                              #     { SENDER -> {
                              #         'hosts': [HOST, HOST, ...]
                              #         'q': { HOST -> QUALITY, HOST -> QUALITY, ... },
                              #         PARAMETER: { HOST -> QUALITY, HOST -> QUALITY, ... },
                              #         ...
                              #     } }
                              #  }
                              # }

        self.routing = {}  # { MEDIUM_NAME ->
                           #    SENDER -> {
                           #        NEXT -> [RECEIVER1, RECEIVER2, ...),
                           #        ...
                           #    }
                           # }
        
    def add_medium(self, name, topology, default_parameters):
        self.mediums[name] = {
            'topology': topology,
            'default_parameters': default_parameters}
                                      # { MEDIUM_NAME -> {
                                      #  'default_parameters' -> {
                                      #   'default_q' -> ...,
                                      #   'default_...' -> ...,
                                      #   'default_...' -> ...,
                                      #  }
                                      #
                                      #  'topology' ->
                                      #     { SENDER -> {
                                      #         'hosts': [HOST, HOST, ...]
                                      #         'q': { HOST -> QUALITY, HOST -> QUALITY, ... }
                                      #         PARAMETER: { HOST -> VALUE, HOST -> VALUE, ... },
                                      #         ...
                                      #     } }
                                      #  }
                                      # }
        self.routing[name] = {}  # SENDER -> {
                                 #     NEXT -> [RECEIVER1, RECEIVER2, ...),
                                 #     ...
                                 # }
        
    def has_medium(self, name):
        return name in self.mediums

    def link_exists(self, channel, sender, receiver):
        """
        Returns True if there exists a link between sender and receiver.
        If communication structure exists the link must be specified. If channel medium has no structure it is assumed
        that channel may be used by all hosts.
        """
        medium_name = channel.tag_name
        if medium_name not in self.mediums:
            return True
        topology = self.mediums[medium_name]['topology']
        if len(topology) == 0:
            return True
        return self.get_link_quality(medium_name, sender, receiver) is not None

    def get_link_parameter_value(self, parameter, medium_name, sender, receiver=None, default=None, no_link_value=None):
        """
        Returns value of parameter between sender and receiver in medium.
        When receiver is None the parameter is get for situation when sender is broadcasting.
        """
        def topology_default():
            defaults = self.mediums[medium_name]['default_parameters']
            default_parameter_name = 'default_'+parameter
            if default_parameter_name in defaults:
                return defaults[default_parameter_name]
            return default
        if medium_name not in self.mediums:
            return default
        medium = self.mediums[medium_name]
        if len(medium['topology']) == 0:
            return default
        if sender not in medium['topology']:
            return no_link_value

        sender_hosts = medium['topology'][sender]['hosts']
        sender_topology = medium['topology'][sender]['parameters']
        if parameter not in sender_topology:
            if receiver in sender_hosts:
                return topology_default()
            else:
                return no_link_value
        sender_parameters = sender_topology[parameter]
        if receiver not in sender_parameters:
            if receiver in sender_hosts:
                return topology_default()
            else:
                return no_link_value
        return sender_parameters[receiver]

    def get_link_quality(self, medium_name, sender, receiver, default=1, no_link_value=None):
        """ """
        return self.get_link_parameter_value('q', medium_name, sender, receiver, default=default,
                                             no_link_value=no_link_value)

    def get_sender_links_qualities(self, medium_name, sender, exclude_broadcast=False):
        """ """
        medium = self.mediums[medium_name]
        if sender not in medium['topology']:
            return {}
        default_quality = 1
        defaults = medium['default_parameters']
        if 'default_q' in defaults:
            default_quality = defaults['default_q']
        sender_topology = medium['topology'][sender]
        sender_qualities = sender_topology['parameters']['q'] if 'q' in sender_topology['parameters'] else {}
        qualities = {}
        for h in sender_topology['hosts']:
            if h is None and exclude_broadcast:
                continue
            if h in sender_qualities:
                qualities[h] = sender_qualities[h]
            else:
                qualities[h] = default_quality
        return qualities

    def _find_existing_next_hop_host(self, topology_name, sender, receiver):
        """
        Finds existing next hop host in the path from sender to receiver.
        If path does not exist, return None.
        """
        routing = self.routing[topology_name]
        if not sender in routing:
            return None
        sender_routing = routing[sender]
        for next_hop in sender_routing:
            if receiver in sender_routing[next_hop]:
                return next_hop
        return None

    def get_hosts_sending_to_receiver(self, medium_name, receiver):
        """
        Returns ditictionary:
         host -> quality
        containing hosts that can send message to receiver.
        """
        if not self.has_medium(medium_name):
            return {}
        hosts = {}
        for sender in self.mediums[medium_name]['topology']:

            q = self.get_link_quality(medium_name, sender, receiver)
            if q is not None:
                hosts[sender] = q
        return hosts

    def get_next_hop_host(self, medium_name, sender, receiver):
        """
        Returns the next host which is in the path between sender and receiver.
        If path does not exist, None is returned.
        """
        # Check if path already exists
        existing_next_hop = self._find_existing_next_hop_host(medium_name, sender, receiver)
        if existing_next_hop is not None:
            return existing_next_hop

        # Build path using Dijsktra algorithm
        topology = self.mediums[medium_name]['topology']

        def find_closest_host(distances, out):
            closest_host = None
            d = -1
            for h in distances:
                if d < 0 or d > distances[h]:
                    if h not in out:
                        closest_host = h
                        d = distances[h]
            return closest_host, d

        # Dijsktra
        distances = {sender: 0}
        out = []
        closest, closes_distance = find_closest_host(distances, out)
        while (closest is not None) and (closest != receiver):

            if closest in topology:
                qualities = self.get_sender_links_qualities(medium_name, closest, exclude_broadcast=True)
                for next_host in qualities:
                    if (next_host not in distances) \
                            or ((closes_distance + qualities[next_host]) < distances[next_host]):
                        distances[next_host] = closes_distance + qualities[next_host]
            out.append(closest)
            closest, closes_distance = find_closest_host(distances, out)

        # for h in distances:
        #     print h.name, distances[h]

        def update_paths(medium_name, receiver, distances):
            routing = self.routing[medium_name]
            # Start from receiver
            current_host = receiver
            distance = distances[receiver]
            # Add receiver to path
            hosts_path = [receiver]
            # Repeat until we finish handling sender
            while distance > 0:
                # Find all hosts that are connected with current host
                previous_hosts = self.get_hosts_sending_to_receiver(medium_name, current_host)
                for prev_host in previous_hosts:
                    # If prev host has not calculated distance, omit him
                    if prev_host not in distances:
                        continue
                    # Check if this host is on the shortest path
                    prev_quality = previous_hosts[prev_host]
                    if distances[prev_host] + prev_quality == distances[current_host]:
                        # Go one step earlier
                        current_host = prev_host
                        # Decrease current distance
                        distance = distances[prev_host]
                        if prev_host not in routing:
                            routing[prev_host] = {}
                        next_host = hosts_path[0]
                        if next_host not in routing[prev_host]:
                            routing[prev_host][next_host] = []
                        for i in range(0, len(hosts_path)):
                            routing[prev_host][next_host].append(hosts_path[i])
                        # Add host to path
                        hosts_path.insert(0, prev_host)
                        break

        # Printer().print_routing(self.routing[medium_name])
        if receiver not in distances:
            raise RuntimeException("The path between {0} and {1} undefined.".format(sender.name, receiver.name))

        update_paths(medium_name, receiver, distances)
        return self._find_existing_next_hop_host(medium_name, sender, receiver)


class Printer():

    def print_topology(self, topology):
        for sender in topology:
            snlen = len(sender.name)
            print sender.name

            for next in topology[sender]['quality']:
                print " " * snlen, ' -> ', next.name, ' : ', topology[sender]['quality'][next]

    def print_routing(self, routing):
        for sender in routing:
            snlen = len(sender.name)
            print sender.name

            for next in routing[sender]:
                nnlen = len(next.name)
                print " " * snlen, ' -> ', next.name

                for receiver in routing[sender][next]:
                    print " " * (snlen+nnlen+4), ' -> ', receiver.name







#!/usr/bin/env python
'''
Created on 28-10-2013

@author: Damian Rusinek <damian.rusinek@gmail.com>
'''

from setuptools import setup, find_packages
import os
import aqopa

def read(filename):
    with open(filename) as f:
        return f.read()

def find_library_data():
    data_files = []
    path = os.path.join('aqopa', 'library', 'models')
    for dir_name in os.listdir(path):
        dir_path = os.path.join(path, dir_name) 
        
        if os.path.isdir(dir_path):
            dir_data_files = []
            for filename in os.listdir(dir_path):
                file_path = os.path.join(dir_path, filename)
                if os.path.isfile(file_path):
                    dir_data_files.append(file_path)
            if len(dir_data_files) > 0:
                data_files.append((dir_path, dir_data_files))
    return data_files
    
setup(name='AQoPA',
      
      version=aqopa.VERSION,
      description='Automated Quality of Protection Analysis Tool for QoP-ML models.',
      long_description=read('README.md'),
      author='Damian Rusinek',
      author_email='damian.rusinek@gmail.com',
      url='http://qopml.org/aqopa/',
      
      license='Freeware',
      platforms=["any"],
      
      install_requires=[read('requirements').split("\n")],
      
      packages=find_packages(),
      package_data={'aqopa' : ['bin/assets/logo.png']},
      data_files=find_library_data(),
      
      entry_points = {
        'console_scripts': [
            'aqopa-gui = aqopa.cmd:gui_command',
            'aqopa-console = aqopa.cmd:console_command',
        ]
      },
      
      classifiers = [
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',
        'Intended Audience :: System Administrators',
        'Intended Audience :: Telecommunications Industry',
        'License :: Freeware',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Security',
        'Topic :: Security :: Cryptography',
        'Topic :: System :: Monitoring',
      ]
     )




#!/usr/bin/env python

import wx
import os
from aqopa.gui.general_purpose_frame_gui import GeneralFrame
from sme.SMETool import SMETool

"""
@file       __init__.py
@brief      gui file for the qopanalysis module
@author     Katarzyna Mazur
"""

class SingleVersionPanel(wx.Panel):
    """
    Frame presenting results for one simulation.
    Simulator may be retrived from module,
    because each module has its own simulator.
    """

    def __init__(self, module, *args, **kwargs):
        wx.Panel.__init__(self, *args, **kwargs)

        self.module = module
        self.versionSimulator = {}
        self.all_facts = []
        self.occured_facts = []

        #################
        # VERSION BOX
        #################

        versionBox = wx.StaticBox(self, label="Version")
        versionsLabel = wx.StaticText(self, label="Choose Version To See\nAnalysis Results:")
        self.versionsList = wx.ComboBox(self, style=wx.TE_READONLY)
        self.versionsList.Bind(wx.EVT_COMBOBOX, self.OnVersionChanged)
        versionBoxSizer = wx.StaticBoxSizer(versionBox, wx.HORIZONTAL)
        versionBoxSizer.Add(versionsLabel, 0, wx.ALL | wx.ALIGN_CENTER, 5)
        versionBoxSizer.Add(wx.StaticText(self), 1, wx.ALL | wx.EXPAND, 5)
        versionBoxSizer.Add(self.versionsList, 1, wx.ALL | wx.ALIGN_CENTER, 5)

        ##################################
        # QoP PARAMETERS BOX
        ##################################
        self.qopParamsBox = wx.StaticBox(self, label="The QoP Analysis Results")
        hostsBox, hostsBoxSizer = self._BuildHostsBoxAndSizer()
        qopParamsBoxSizer = wx.StaticBoxSizer(self.qopParamsBox, wx.VERTICAL)
        qopParamsBoxSizer.Add(hostsBoxSizer, 1, wx.ALL | wx.EXPAND, 5)

        #################
        # BUTTONS LAY
        #################
        self.showQoPBtn = wx.Button(self, label="Show")
        self.showQoPBtn.Bind(wx.EVT_BUTTON, self.OnShowQoPBtnClicked)
        #self.launchSMEBtn = wx.Button(self, label="Evaluate")
        #self.launchSMEBtn.Bind(wx.EVT_BUTTON, self.OnLaunchSMEClicked)
        buttonsSizer = wx.BoxSizer(wx.HORIZONTAL)
        buttonsSizer.Add(wx.StaticText(self), 1, wx.ALL | wx.EXPAND, 5)
        buttonsSizer.Add(self.showQoPBtn, 0, wx.ALL | wx.EXPAND, 5)
        #buttonsSizer.Add(self.launchSMEBtn, 0, wx.ALL | wx.EXPAND, 5)

        #################
        # MAIN LAY
        #################
        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(versionBoxSizer, 0, wx.ALL | wx.EXPAND, 5)
        sizer.Add(qopParamsBoxSizer, 0, wx.ALL | wx.EXPAND, 5)
        sizer.Add(buttonsSizer, 0, wx.ALL | wx.EXPAND, 5)
        self.SetSizer(sizer)

        self.SetVersionsResultsVisibility(False)

    def OnShowQoPBtnClicked(self, event):
        hostName = self.hostsList.GetValue()
        if hostName != "" :
            versionName = self.versionsList.GetValue()
            simulator = self.versionSimulator[versionName]
            host = self._GetSelectedHost(simulator)
            self.ShowQoPParameters(simulator, host)

    def OnLaunchSMEClicked(self, event):
        smetool = SMETool(None)
        smetool.SetClientSize(wx.Size(800,450))
        smetool.CentreOnScreen()
        smetool.Show()

    def _GetSelectedHost(self, simulator):

        host = None

        # get selected module name from combo
        hostName = self.hostsList.GetValue()

        # find host with the selected name
        for h in simulator.context.hosts:
            if h.original_name() == hostName:
                host = h
                break

        return host

    def _PopulateComboWithHostsNames(self, simulator):
        hostsNames = []
        [hostsNames.append(h.original_name()) for h in simulator.context.hosts if h.original_name() not in hostsNames]
        self.hostsList.Clear()
        self.hostsList.AppendItems(hostsNames)

    #################
    # LAYOUT
    #################

    def _BuildHostsBoxAndSizer(self):
        """ """

        self.chooseHostLbl = wx.StaticText(self, label="Choose Host To See\nit's QoP Parameters:")
        self.hostsList = wx.ComboBox(self, style=wx.TE_READONLY)

        self.hostsBox = wx.StaticBox(self, label="Host(s)")
        self.hostsBoxSizer = wx.StaticBoxSizer(self.hostsBox, wx.HORIZONTAL)

        self.hostsBoxSizer.Add(self.chooseHostLbl, 0, wx.ALL | wx.EXPAND, 5)
        self.hostsBoxSizer.Add(wx.StaticText(self), 1, wx.ALL | wx.EXPAND, 5)
        self.hostsBoxSizer.Add(self.hostsList, 1, wx.ALL | wx.EXPAND | wx.ALIGN_RIGHT, 5)

        return self.hostsBox, self.hostsBoxSizer

    #################
    # REACTIONS
    #################

    def AddFinishedSimulation(self, simulator):
        """ """
        version = simulator.context.version
        self.versionsList.Append(version.name)
        self.versionSimulator[version.name] = simulator

    def OnVersionChanged(self, event):
        """ """
        versionName = self.versionsList.GetValue()
        simulator = self.versionSimulator[versionName]
        self._PopulateComboWithHostsNames(simulator)
        self.SetVersionsResultsVisibility(True)

    def RemoveAllSimulations(self):
        """ """
        self.versionsList.Clear()
        self.versionsList.SetValue("")
        self.versionSimulator = {}
        self.hostsList.Clear()
        self.hostsList.SetValue("")
        self.SetVersionsResultsVisibility(False)

    def SetVersionsResultsVisibility(self, visible):
        """ """
        widgets = []
        widgets.append(self.hostsList)
        widgets.append(self.hostsBox)
        widgets.append(self.showQoPBtn)
        #widgets.append(self.launchSMEBtn)
        widgets.append(self.qopParamsBox)
        widgets.append(self.chooseHostLbl)

        for w in widgets:
            if visible:
                w.Show()
            else:
                w.Hide()

        self.Layout()

    def ShowQoPParameters(self, simulator, host) :

        title = "QoP Parameters for host: "
        title += host.original_name()

        qopsWindow = GeneralFrame(self, "QoP Analysis Results", title, "modules_results.png")
        qopParamsPanel = wx.Panel(qopsWindow)

        # get all & occured facts
        versionName = self.versionsList.GetValue()
        simulator = self.versionSimulator[versionName]
        # simply copy lists
        self.occured_facts = self._GetoOccuredFacts(simulator)[:]
        self.all_facts = self._GetAllFacts(simulator)[:]

        ##################################
        # ALL FACTS LAYOUT
        ##################################
        all_factsListBox = wx.ListBox(qopParamsPanel, choices=self.all_facts)
        all_factsBox = wx.StaticBox(qopParamsPanel, label="All Facts")
        allFactsBoxSizer = wx.StaticBoxSizer(all_factsBox, wx.VERTICAL)
        allFactsBoxSizer.Add(all_factsListBox, 1, wx.ALL | wx.EXPAND, 5)

        ##################################
        # OCCURED FACTS LAYOUT
        ##################################
        occured_factsListBox = wx.ListBox(qopParamsPanel, choices=self.occured_facts)
        occured_factsBox = wx.StaticBox(qopParamsPanel, label="Occured Facts")
        occuredFactsBoxSizer = wx.StaticBoxSizer(occured_factsBox, wx.VERTICAL)
        occuredFactsBoxSizer.Add(occured_factsListBox, 1, wx.ALL | wx.EXPAND, 5)

        sizer = wx.BoxSizer(wx.HORIZONTAL)
        sizer.Add(allFactsBoxSizer, 1, wx.ALL | wx.EXPAND, 5)
        sizer.Add(occuredFactsBoxSizer, 1, wx.ALL | wx.EXPAND, 5)

        qopParamsPanel.SetSizer(sizer)
        qopParamsPanel.Layout()

        qopsWindow.CentreOnScreen()
        qopsWindow.AddPanel(qopParamsPanel)
        qopsWindow.SetWindowSize(600, 300)
        qopsWindow.Show()

        # some kind of debugging
        #print "Occured facts from GUI: "+str(self._GetoOccuredFacts(simulator))
        #print "All facts from GUI: "+str(self._GetAllFacts(simulator))

    def _GetAllFacts(self, simulator):
        return self.module.get_all_facts()

    def _GetoOccuredFacts(self, simulator):
        host = None
        # get all hosts assigned to this simulator
        allHosts = self.module.occured_facts[simulator]
        # get the name of the host selected on hosts combo box
        hostName = self.hostsList.GetValue()
        # from all hosts get the selected one - its the host
        # with the same same selected on combobox
        for h in allHosts :
            if h.original_name() == hostName :
                host = h
                break
        # get all facts for the particular simulator and host
        return self.module.get_occured_facts(simulator, host)

class MainResultsNotebook(wx.Notebook):
    """ """
    def __init__(self, module, *args, **kwargs):
        wx.Notebook.__init__(self, *args, **kwargs)

        self.module = module

        il = wx.ImageList(24, 24)
        singleVersionImg = il.Add(wx.Bitmap(self.CreatePath4Resource('qop.png'), wx.BITMAP_TYPE_PNG))
        self.AssignImageList(il)

        self.oneVersionTab = SingleVersionPanel(self.module, self)
        self.AddPage(self.oneVersionTab, "Single Version")
        self.SetPageImage(0, singleVersionImg)
        self.oneVersionTab.Layout()

    def OnParsedModel(self):
        """ """
        self.oneVersionTab.RemoveAllSimulations()

    def OnSimulationFinished(self, simulator):
        """ """
        self.oneVersionTab.AddFinishedSimulation(simulator)

    def OnAllSimulationsFinished(self, simulators):
        """ """
        pass

    def CreatePath4Resource(self, resourceName):
        """
        @brief      creates and returns path to the
                    given file in the resource
                    ('assets') dir
        @return     path to the resource
        """
        tmp = os.path.split(os.path.dirname(__file__))
        # find last / character in path
        idx = tmp[0].rfind('/')
        # get substring - path for resource
        path = tmp[0][0:idx]
        return os.path.join(path, 'bin', 'assets', resourceName)

class ModuleGui(wx.EvtHandler):
    def __init__(self, module):
        """ """
        wx.EvtHandler.__init__(self)
        self.module = module
        self.mainResultNotebook = None

    def get_gui(self):
        if not getattr(self, '__gui', None):
            setattr(self, '__gui', ModuleGui(self))
        return getattr(self, '__gui', None)

    def get_name(self):
        return "QoP Analysis"

    def install_gui(self, simulator):
        """ Install module for gui simulation """
        self._install(simulator)
        return simulator

    def get_configuration_panel(self, parent):
        """ Returns WX panel with configuration controls. """

        panel = wx.Panel(parent)
        sizer = wx.BoxSizer(wx.VERTICAL)
        text = wx.StaticText(panel, label="Module does not need to be configured.")
        sizer.Add(text, 0, wx.ALL | wx.EXPAND, 5)
        text = wx.StaticText(panel, label="All result options will be available after results are calculated.")
        sizer.Add(text, 0, wx.ALL | wx.EXPAND, 5)
        panel.SetSizer(sizer)
        return panel

    def get_results_panel(self, parent):
        """
        Create main result panel existing from the beginning
        which will be extended when versions' simulations are finished.
        """
        self.mainResultNotebook = MainResultsNotebook(self.module, parent)
        return self.mainResultNotebook

    def on_finished_simulation(self, simulator):
        """ """
        self.mainResultNotebook.OnSimulationFinished(simulator)

    def on_finished_all_simulations(self, simulators):
        """
        Called once for all simulations after all of them are finished.
        """
        self.mainResultNotebook.OnAllSimulationsFinished(simulators)

    def on_parsed_model(self):
        """ """
        self.mainResultNotebook.OnParsedModel()